"""Transformations between representations."""

__all__: list[str] = []

from typing import Any
from warnings import warn

from plum import dispatch

import quaxed.numpy as jnp
from unxt.quantity import AbstractQuantity, Quantity

from coordinax._src.vectors.d1 import CartesianPos1D, RadialPos
from coordinax._src.vectors.d2 import AbstractPos2D, CartesianPos2D, PolarPos
from coordinax._src.vectors.d3 import (
    AbstractPos3D,
    CartesianPos3D,
    CylindricalPos,
    MathSphericalPos,
    SphericalPos,
)
from coordinax._src.vectors.exceptions import IrreversibleDimensionChange


@dispatch.multi(
    (type[CylindricalPos], CartesianPos2D),
    (type[SphericalPos], CartesianPos2D),
    (type[MathSphericalPos], CartesianPos2D),
)
def vconvert(
    target: type[AbstractPos3D],
    current: AbstractPos2D,
    /,
    z: AbstractQuantity = Quantity(0, "m"),
    **kwargs: Any,
) -> AbstractPos3D:
    """AbstractPos2D -> Cartesian2D -> Cartesian3D -> AbstractPos3D.

    The 2D vector is in the xy plane. The `z` coordinate is a keyword argument and
    defaults to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.CartesianPos2D.from_([1, 2], "km")

    >>> x2 = cx.vconvert(cx.vecs.CylindricalPos, x, z=u.Quantity(14, "km"))
    >>> print(x2)
    <CylindricalPos (rho[km], phi[rad], z[km])
        [ 2.236  1.107 14.   ]>

    >>> x3 = cx.vconvert(cx.SphericalPos, x, z=u.Quantity(14, "km"))
    >>> x3
    SphericalPos( r=Distance(value=f32[], unit=Unit("km")),
                  theta=Angle(value=f32[], unit=Unit("rad")),
                  phi=Angle(value=f32[], unit=Unit("rad")) )
    >>> x3.r
    Distance(Array(14.177447, dtype=float32), unit='km')

    >>> x3 = cx.vconvert(cx.vecs.MathSphericalPos, x, z=u.Quantity(14, "km"))
    >>> x3
    MathSphericalPos( r=Distance(value=f32[], unit=Unit("km")),
                      theta=Angle(value=f32[], unit=Unit("rad")),
                      phi=Angle(value=f32[], unit=Unit("rad")) )
    >>> x3.r
    Distance(Array(14.177447, dtype=float32), unit='km')

    """
    cart2 = vconvert(CartesianPos2D, current)
    cart3 = vconvert(CartesianPos3D, cart2, z=z)
    return vconvert(target, cart3)


@dispatch.multi(
    (type[CartesianPos3D], PolarPos),
)
def vconvert(
    target: type[AbstractPos3D],
    current: AbstractPos2D,
    /,
    z: AbstractQuantity = Quantity(0, "m"),
    **kwargs: Any,
) -> AbstractPos3D:
    """AbstractPos2D -> PolarPos -> Cylindrical -> AbstractPos3D.

    The 2D vector is in the xy plane. The `z` coordinate is a keyword argument and
    defaults to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.PolarPos(r=Quantity(1, "km"), phi=u.Quantity(10, "deg"))

    >>> x2 = cx.vconvert(cx.CartesianPos3D, x, z=u.Quantity(14, "km"))
    >>> print(x2)
    <CartesianPos3D (x[km], y[km], z[km])
        [ 0.985  0.174 14.   ]>

    """
    polar = vconvert(PolarPos, current)
    cyl = vconvert(CylindricalPos, polar, z=z)
    return vconvert(target, cyl)


# =============================================================================
# CartesianPos2D


# -----------------------------------------------
# 1D


@dispatch
def vconvert(
    target: type[CartesianPos1D], current: CartesianPos2D, /, **kwargs: Any
) -> CartesianPos1D:
    """CartesianPos2D -> CartesianPos1D.

    The `y` coordinate is dropped.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.CartesianPos2D.from_([1, 2], "km")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.vconvert(cx.vecs.CartesianPos1D, x, z=u.Quantity(14, "km"))
    >>> print(x2)
    <CartesianPos1D (x[km])
        [1]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.x)


@dispatch
def vconvert(
    target: type[RadialPos], current: CartesianPos2D, /, **kwargs: Any
) -> RadialPos:
    """CartesianPos2D -> RadialPos.

    The `x` and `y` coordinates are converted to the radial coordinate `r`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.CartesianPos2D.from_([1, 2], "km")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.vconvert(cx.vecs.RadialPos, x, z=u.Quantity(14, "km"))
    >>> print(x2)
    <RadialPos (r[km])
        [2.236]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=jnp.sqrt(current.x**2 + current.y**2))


# -----------------------------------------------
# 3D


@dispatch
def vconvert(
    target: type[CartesianPos3D],
    current: CartesianPos2D,
    /,
    *,
    z: AbstractQuantity = Quantity(0, "m"),
    **kwargs: Any,
) -> CartesianPos3D:
    """CartesianPos2D -> CartesianPos3D.

    The `x` and `y` coordinates are converted to the `x` and `y` coordinates of
    the 3D system.  The `z` coordinate is a keyword argument and defaults to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.CartesianPos2D.from_([1, 2], "km")

    >>> x2 = cx.vconvert(cx.CartesianPos3D, x, z=u.Quantity(14, "km"))
    >>> print(x2)
    <CartesianPos3D (x[km], y[km], z[km])
        [ 1  2 14]>

    """
    return target(x=current.x, y=current.y, z=z)


# =============================================================================
# PolarPos

# -----------------------------------------------
# 1D


@dispatch
def vconvert(
    target: type[CartesianPos1D], current: PolarPos, /, **kwargs: Any
) -> CartesianPos1D:
    """PolarPos -> CartesianPos1D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.PolarPos(r=u.Quantity(1, "km"), phi=u.Quantity(10, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.vconvert(cx.vecs.CartesianPos1D, x)
    >>> print(x2)
    <CartesianPos1D (x[km])
        [0.985]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.r * jnp.cos(current.phi))


@dispatch
def vconvert(target: type[RadialPos], current: PolarPos, /, **kwargs: Any) -> RadialPos:
    """PolarPos -> RadialPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.PolarPos(r=u.Quantity(1, "km"), phi=u.Quantity(10, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.vconvert(cx.vecs.RadialPos, x)
    >>> print(x2)
    <RadialPos (r[km])
        [1]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=current.r)


# -----------------------------------------------
# 3D


@dispatch
def vconvert(
    target: type[SphericalPos],
    current: PolarPos,
    /,
    theta: Quantity["angle"] = Quantity(0, "radian"),
    **kwargs: Any,
) -> SphericalPos:
    """PolarPos -> SphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.PolarPos(r=Quantity(1, "km"), phi=u.Quantity(10, "deg"))

    >>> x2 = cx.vconvert(cx.SphericalPos, x, theta=u.Quantity(14, "deg"))
    >>> print(x2)
    <SphericalPos (r[km], theta[deg], phi[deg])
        [ 1 14 10]>

    """
    return target.from_(r=current.r, theta=theta, phi=current.phi)


@dispatch
def vconvert(
    target: type[MathSphericalPos],
    current: PolarPos,
    /,
    phi: Quantity["angle"] = Quantity(0, "radian"),
    **kwargs: Any,
) -> MathSphericalPos:
    """PolarPos -> MathSphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.PolarPos(r=u.Quantity(1, "km"), phi=u.Quantity(10, "deg"))

    >>> x2 = cx.vconvert(cx.vecs.MathSphericalPos, x, phi=u.Quantity(14, "deg"))
    >>> print(x2)
    <MathSphericalPos (r[km], theta[deg], phi[deg])
        [ 1 10 14]>

    """
    return target.from_(r=current.r, phi=phi, theta=current.phi)


@dispatch
def vconvert(
    target: type[CylindricalPos],
    current: PolarPos,
    /,
    *,
    z: Quantity["length"] = Quantity(0, "m"),
    **kwargs: Any,
) -> CylindricalPos:
    """PolarPos -> CylindricalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.PolarPos(r=u.Quantity(1, "km"), phi=u.Quantity(10, "deg"))

    >>> x2 = cx.vconvert(cx.vecs.CylindricalPos, x, z=u.Quantity(14, "km"))
    >>> print(x2)
    <CylindricalPos (rho[km], phi[deg], z[km])
        [ 1 10 14]>

    """
    return target(rho=current.r, phi=current.phi, z=z)



"""Transformations between representations."""

__all__: list[str] = []


from plum import dispatch

import quaxed.numpy as jnp

from coordinax._src.vectors.d3 import CylindricalPos, CylindricalVel
from coordinax._src.vectors.dn import PoincarePolarVector
from coordinax._src.vectors.space import Space


@dispatch
def vconvert(target: type[Space], w: Space, /) -> Space:
    """Space -> Space.

    Examples
    --------
    >>> import coordinax as cx

    >>> w = cx.Space(
    ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
    ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s")
    ... )

    >>> cx.vconvert(cx.Space, w)
    Space({ 'length': CartesianPos3D( ... ),
             'speed': CartesianVel3D( ... ) })

    """
    return w


@dispatch
def vconvert(target: type[PoincarePolarVector], w: Space, /) -> PoincarePolarVector:
    """Space -> PoincarePolarVector.

    Examples
    --------
    >>> import coordinax as cx

    >>> w = cx.Space(
    ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
    ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s")
    ... )

    >>> cx.vconvert(cx.vecs.PoincarePolarVector, w)
    PoincarePolarVector(
      rho=Quantity[...](value=f32[1,2], unit=Unit("m")),
      pp_phi=Quantity[...]( value=f32[1,2], unit=Unit("m rad(1/2) / s(1/2)") ),
      z=Quantity[...](value=i32[1,2], unit=Unit("m")),
      d_rho=Quantity[...]( value=f32[1,2], unit=Unit("m / s") ),
      d_pp_phi=Quantity[...]( value=f32[1,2], unit=Unit("m rad(1/2) / s(1/2)") ),
      d_z=Quantity[...]( value=f32[1,2], unit=Unit("m / s") )
    )

    """
    q = w["length"].vconvert(CylindricalPos)
    p = w["speed"].vconvert(CylindricalVel, q)

    # pg. 437, Papaphillipou & Laskar (1996)
    sqrt2theta = jnp.sqrt(jnp.abs(2 * q.rho**2 * p.d_phi))
    pp_phi = sqrt2theta * jnp.cos(q.phi)
    pp_phidot = sqrt2theta * jnp.sin(q.phi)

    return PoincarePolarVector(
        rho=q.rho, pp_phi=pp_phi, z=q.z, d_rho=p.d_rho, d_pp_phi=pp_phidot, d_z=p.d_z
    )



# ruff: noqa: ERA001
"""Galilean coordinate transformations."""

__all__ = ["GalileanOperator"]


from typing import TYPE_CHECKING, Any, final, overload

import equinox as eqx
from plum import dispatch

import quaxed.numpy as xp
from dataclassish.converters import Unless

from .base import AbstractGalileanOperator
from .boost import GalileanBoost
from .rotation import GalileanRotation
from .spatial_translation import GalileanSpatialTranslation
from .translation import GalileanTranslation
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.operators.composite import AbstractCompositeOperator
from coordinax._src.operators.identity import Identity
from coordinax._src.operators.pipe import Pipe

if TYPE_CHECKING:
    from typing import Self


@final
class GalileanOperator(AbstractCompositeOperator, AbstractGalileanOperator):
    r"""Operator for general Galilean transformations.

    In the transformed frame the coordinates are given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t+s, R \mathbf{x} + \mathbf{v} t + \mathbf{a})

    where :math:`R` is the rotation matrix, :math:`\mathbf{v}` is the boost
    velocity, :math:`\mathbf{a}` is the spatial translationl, and :math:`s` is
    the time translation. This is equivalent to a sequential operation of 1. a
    rotation, 2. a translation, 3. a boost.

    Parameters
    ----------
    translation : `coordinax.ops.GalileanTranslation`
        The spatial translation of the frame. See
        :class:`coordinax.ops.GalileanTranslation` for alternative
        inputs to construct this parameter.
    velocity : :class:`coordinax.ops.GalileanBoost`
        The boost to the frame. See
        :class:`coordinax.ops.GalileanBoost` for alternative
        inputs to construct this parameter.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanOperator(
    ...     translation=u.Quantity([0., 2., 3., 4.], "km"),
    ...     velocity=u.Quantity([1., 2., 3.], "km/s"))
    >>> op
    GalileanOperator(
      rotation=GalileanRotation(rotation=f32[3,3]),
      translation=GalileanTranslation(
        translation=FourVector(
          t=Quantity[...](value=f32[], unit=Unit("s")),
          q=CartesianPos3D( ... ) )
      ),
      velocity=GalileanBoost( velocity=CartesianVel3D( ... ) )
    )

    Note that the translation is a
    :class:`coordinax.ops.GalileanTranslation` with a
    :class:`vector.FourVector` translation, and the velocity is a
    :class:`coordinax.ops.GalileanBoost` with a
    :class:`coordinax.vecs.AbstractVel` velocity. We can also construct them
    directly, which allows for other vector types.

    >>> op = cx.ops.GalileanOperator(
    ...     translation=cx.ops.GalileanTranslation(
    ...         cx.FourVector(t=u.Quantity(2.5, "Gyr"),
    ...                       q=cx.SphericalPos(r=u.Quantity(1, "km"),
    ...                                         theta=u.Quantity(90, "deg"),
    ...                                         phi=u.Quantity(0, "rad") ) ) ),
    ...     velocity=cx.ops.GalileanBoost(
    ...         cx.CartesianVel3D.from_([1, 2, 3], "km/s") )
    ... )
    >>> op
    GalileanOperator(
      rotation=GalileanRotation(rotation=f32[3,3]),
      translation=GalileanTranslation(
        translation=FourVector(
          t=Quantity[...)](value=...f32[], unit=Unit("Gyr")),
          q=SphericalPos( ... )
        )
      ),
      velocity=GalileanBoost( velocity=CartesianVel3D( ... ) )
    )

    Galilean operators can be applied to :class:`vector.FourVector`:

    >>> w = cx.FourVector.from_([0, 0, 0, 0], "km")
    >>> new = op(w)
    >>> new
    FourVector(
      t=Quantity[...](value=...f32[], unit=Unit("s")),
      q=CartesianPos3D( ... )
    )
    >>> new.t.ustrip("Gyr").round(2)
    Array(2.5, dtype=float32, ...)
    >>> print(new.q)
    <CartesianPos3D (x[km], y[km], z[km])
        [7.889e+16 1.578e+17 2.367e+17]>

    Also the Galilean operators can also be applied to
    :class:`vector.AbstractPos3D` and :class:`unxt.Quantity`:

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "km")
    >>> t = u.Quantity(0, "s")
    >>> newq, newt = op(q, t)
    >>> print(newq)
    <CartesianPos3D (x[km], y[km], z[km])
        [7.889e+16 1.578e+17 2.367e+17]>

    >>> newt
    Quantity['time'](Array(7.8894005e+16, dtype=float32, ...), unit='s')

    """

    rotation: GalileanRotation = eqx.field(
        default=GalileanRotation(xp.eye(3)),
        converter=GalileanRotation.from_,
    )
    """The in-frame spatial rotation."""

    translation: GalileanTranslation = eqx.field(
        default=GalileanTranslation.from_([0, 0, 0, 0], "km"),
        converter=Unless(GalileanTranslation, converter=GalileanTranslation.from_),
    )
    """The temporal + spatial translation.

    The translation vector [T, Q].  This parameters accetps either a
    :class:`coordinax.ops.GalileanTranslation` instance or
    any input that can be used to construct a :meth:`vector.FourVector`, using
    :meth:`vector.FourVector.from_`. See :class:`vector.FourVector` for
    details.
    """

    velocity: GalileanBoost = eqx.field(
        default=GalileanBoost.from_([0, 0, 0], "km/s"),
        converter=Unless(GalileanBoost, converter=GalileanBoost.from_),
    )
    """The boost to the frame.

    This parameters accepts either a :class:`coordinax.ops.GalileanBoost`
    instance or any input that can be used to construct one. See
    :class:`coordinax.ops.GalileanBoost.from_` for details.
    """

    @property
    def operators(
        self,
    ) -> tuple[GalileanRotation, GalileanTranslation, GalileanBoost]:
        """Rotation -> translation -> boost."""
        return (self.rotation, self.translation, self.velocity)

    @overload
    def __getitem__(self, key: int) -> AbstractOperator: ...

    @overload
    def __getitem__(self, key: slice) -> "Self": ...

    def __getitem__(self, key: int | slice) -> "AbstractOperator | Self":
        """Getitem from the operators.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> op = cx.ops.GalileanOperator(
        ...     translation=u.Quantity([0., 2., 3., 4.], "km"),
        ...     velocity=u.Quantity([1., 2., 3.], "km/s"))

        >>> op[0]
        GalileanRotation(rotation=f32[3,3])

        >>> op[1:]
        Pipe(( GalileanTranslation(FourVector( ... )),
                   GalileanBoost(CartesianVel3D( ... )) ))

        """
        if isinstance(key, int):
            return self.operators[key]
        return Pipe(self.operators[key])


@dispatch
def simplify_op(
    op: GalileanOperator, /, **kwargs: Any
) -> (
    GalileanOperator
    | Pipe
    | GalileanBoost
    | GalileanRotation
    | GalileanTranslation
    | GalileanSpatialTranslation
    | Identity
):
    """Simplify a Galilean operator.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    This Galilean operator cannot be simplified:

    >>> op = cx.ops.GalileanOperator(
    ...     translation=u.Quantity([0., 2., 3., 4.], "km"),
    ...     velocity=u.Quantity([1., 2., 3.], "km/s"),
    ...     rotation=jnp.eye(3).at[0, 2].set(1),
    ... )
    >>> op
    GalileanOperator(
      rotation=GalileanRotation(rotation=f32[3,3]),
      translation=GalileanTranslation(
        translation=FourVector(
          t=Quantity[...](value=f32[], unit=Unit("s")),
          q=CartesianPos3D( ... ) )
      ),
      velocity=GalileanBoost( velocity=CartesianVel3D( ... ) )
    )

    >>> cx.ops.simplify_op(op) is op
    True

    This Galilean operator can be simplified in all its components except the
    translation:

    >>> op = cx.ops.GalileanOperator(translation=u.Quantity([0., 2., 3., 4.], "km"))
    >>> cx.ops.simplify_op(op)
    GalileanTranslation(FourVector( ... ))

    """
    simple_ops = [simplify_op(x, **kwargs) for x in op.operators]
    if any(
        not isinstance(x, type(orig))
        for x, orig in zip(simple_ops, op.operators, strict=True)
    ):
        return Pipe(simple_ops).simplify()

    return op



"""Galilean coordinate transformations."""
# ruff: noqa: ERA001, N806

__all__ = ["GalileanRotation"]


from collections.abc import Mapping
from dataclasses import replace
from typing import Any, Literal, TypeAlias, final

import equinox as eqx
import jax
from jax.scipy.spatial.transform import Rotation
from jaxtyping import Array, Shaped
from plum import convert, dispatch
from quax import quaxify

import quaxed.numpy as jnp
import unxt as u

from .base import AbstractGalileanOperator
from coordinax._src.angles import Angle
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.operators.identity import Identity
from coordinax._src.vectors.base import ToUnitsOptions
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.base_vel import AbstractVel
from coordinax._src.vectors.d3 import AbstractPos3D

vec_matmul = quaxify(jax.numpy.vectorize(jax.numpy.matmul, signature="(3,3),(3)->(3)"))

RotationMatrix: TypeAlias = Shaped[Array, "3 3"]


def converter(x: Any) -> Array:
    """Convert the input to a rotation matrix."""
    if isinstance(x, GalileanRotation):
        out = x.rotation
    elif isinstance(x, u.Quantity):
        out = u.ustrip("", x)
    else:
        out = x
    return jnp.asarray(out)


@final
class GalileanRotation(AbstractGalileanOperator):
    r"""Operator for Galilean rotations.

    The coordinate transform is given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t, R \mathbf{x})

    where :math:`R` is the rotation matrix.  Note this is NOT time dependent.

    Parameters
    ----------
    rotation : Array[float, (3, 3)]
        The rotation matrix.

    Raises
    ------
    ValueError
        If the rotation matrix is not orthogonal.

    Notes
    -----
    The Galilean rotation is not a time-dependent transformation.  This is part
    of the inhomogeneous Galilean group, which is the group of transformations
    that leave the space-time interval invariant.

    Examples
    --------
    We start with the required imports:

    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    We can then create a rotation operator:

    >>> Rz = jnp.asarray([[0, -1, 0], [1, 0,  0], [0, 0, 1]])
    >>> op = cx.ops.GalileanRotation(Rz)
    >>> op
    GalileanRotation(rotation=i32[3,3])

    Translation operators can be applied to a Quantity[float, (N, 3), "...]:

    >>> q = u.Quantity([1, 0, 0], "m")
    >>> t = u.Quantity(1, "s")
    >>> newq, newt = op(q, t)
    >>> newq
    Quantity['length'](Array([0, 1, 0], dtype=int32), unit='m')

    The time is not affected by the rotation.

    >>> newt
    Quantity['time'](Array(1, dtype=int32, ...), unit='s')

    This also works for a batch of vectors:

    >>> q = u.Quantity([[1, 0, 0], [0, 1, 0]], "m")
    >>> t = u.Quantity(0, "s")

    >>> newq, newt = op(q, t)
    >>> newq
    Quantity['length'](Array([[ 0,  1,  0],
                              [-1,  0,  0]], dtype=int32), unit='m')

    Translation operators can be applied to :class:`vector.AbstractPos3D`:

    >>> q = cx.CartesianPos3D.from_(q)  # from the previous example
    >>> newq, newt = op(q, t)
    >>> newq.x
    Quantity['length'](Array([ 0, -1], dtype=int32), unit='m')
    >>> newq.norm().value.round(2)
    Array([1., 1.], dtype=float32)

    """

    rotation: Shaped[Array, "3 3"] = eqx.field(converter=converter)
    """The rotation vector."""

    #: Tolerance check on the rotation matrix.
    check_tol: Mapping[str, Any] = eqx.field(
        default_factory=lambda: {"atol": 1e-7}, repr=False, static=True
    )

    # TODO: fix so that it works in jitted contexts
    # def __check_init__(self) -> None:
    #     # Check that the rotation matrix is orthogonal.
    #     _ = eqx.error_if(
    #         self.rotation,
    #         not jnp.allclose(
    #             self.rotation @ self.rotation.T, jnp.eye(3), **self.check_tol
    #         ),
    #         "The rotation matrix must be orthogonal.",
    #     )

    # -----------------------------------------------------

    @classmethod
    def from_euler(
        cls: "type[GalileanRotation]", seq: str, angles: u.Quantity["angle"] | Angle, /
    ) -> "GalileanRotation":
        """Initialize from Euler angles.

        See `jax.scipy.spatial.transform.Rotation.from_euler`.
        `XYZ` are intrinsic rotations, `xyz` are extrinsic rotations.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> op = cx.ops.GalileanRotation.from_euler("z", u.Quantity(90, "deg"))
        >>> op.rotation.round(2)
        Array([[ 0., -1.,  0.],
               [ 1.,  0.,  0.],
               [ 0.,  0.,  1.]], dtype=float32)

        """
        R = Rotation.from_euler(seq, u.ustrip("deg", angles), degrees=True).as_matrix()

        return cls(rotation=R)

    @classmethod
    @AbstractOperator.from_.dispatch  # type: ignore[attr-defined, misc]
    def from_(cls: "type[GalileanRotation]", obj: Rotation, /) -> "GalileanRotation":
        """Initialize from a `jax.scipy.spatial.transform.Rotation`.

        Examples
        --------
        >>> import jax.numpy as jnp
        >>> from jax.scipy.spatial.transform import Rotation
        >>> import coordinax as cx

        >>> R = Rotation.from_euler("z", 90, degrees=True)
        >>> op = cx.ops.GalileanRotation.from_(R)

        >>> jnp.allclose(op.rotation, R.as_matrix())
        Array(True, dtype=bool)

        """
        return cls(rotation=obj.as_matrix())

    # -----------------------------------------------------

    @property
    def is_inertial(self) -> Literal[True]:
        """Galilean rotation is an inertial-frame preserving transform.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> import coordinax as cx

        >>> Rz = jnp.asarray([[0, -1, 0], [1, 0,  0], [0, 0, 1]])
        >>> op = cx.ops.GalileanRotation(Rz)
        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "GalileanRotation":
        """The inverse of the operator.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> import coordinax as cx

        >>> Rz = jnp.asarray([[0, -1, 0], [1, 0,  0], [0, 0, 1]])
        >>> op = cx.ops.GalileanRotation(Rz)
        >>> op.inverse
        GalileanRotation(rotation=i32[3,3])

        >>> jnp.allclose(op.rotation, op.inverse.rotation.T)
        Array(True, dtype=bool)

        """
        return replace(self, rotation=self.rotation.T)

    # -----------------------------------------------------

    @AbstractOperator.__call__.dispatch(precedence=1)
    def __call__(
        self: "GalileanRotation",
        q: Shaped[u.Quantity["length"], "*batch 3"],
        /,
        **__: Any,
    ) -> Shaped[u.Quantity["length"], "*batch 3"]:
        """Apply the rotation to the coordinates.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> import unxt as u
        >>> import coordinax as cx

        >>> Rz = jnp.asarray([[0, -1, 0], [1, 0,  0], [0, 0, 1]])
        >>> op = cx.ops.GalileanRotation(Rz)

        >>> q = u.Quantity([1, 0, 0], "m")
        >>> op(q)
        Quantity['length'](Array([0, 1, 0], dtype=int32), unit='m')

        THere's a related dispatch that also takes a time argument:

        >>> t = u.Quantity(1, "s")
        >>> newq, newt = op(q, t)
        >>> newq
        Quantity['length'](Array([0, 1, 0], dtype=int32), unit='m')

        The time is not affected by the rotation.
        >>> newt
        Quantity['time'](Array(1, dtype=int32, ...), unit='s')

        """
        return vec_matmul(self.rotation, q)

    @AbstractOperator.__call__.dispatch
    def __call__(
        self: "GalileanRotation", q: AbstractPos3D, /, **__: Any
    ) -> AbstractPos3D:
        """Apply the rotation to the coordinates.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> import unxt as u
        >>> import coordinax as cx

        >>> Rz = jnp.asarray([[0, -1, 0], [1, 0,  0], [0, 0, 1]])
        >>> op = cx.ops.GalileanRotation(Rz)

        >>> q = cx.CartesianPos3D.from_([1, 0, 0], "m")
        >>> newq = op(q)
        >>> newq.x
        Quantity['length'](Array(0, dtype=int32), unit='m')

        """
        return self.rotation @ q

    # -----------------------------------------------------
    # Arithmetic operations

    def __neg__(self: "GalileanRotation") -> "GalileanRotation":
        """Negate the rotation.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> import coordinax as cx

        >>> Rz = jnp.asarray([[0, -1, 0], [1, 0,  0], [0, 0, 1]])
        >>> op = cx.ops.GalileanRotation(Rz)
        >>> print((-op).rotation)
        [[ 0  1  0]
         [-1  0  0]
         [ 0  0 -1]]

        """
        return replace(self, rotation=-self.rotation)

    @dispatch.abstract
    def __matmul__(self: "GalileanRotation", other: Any, /) -> Any: ...


# ============================================================================
# Call dispatches


@AbstractOperator.__call__.dispatch
def call(
    self: GalileanRotation, q: AbstractPos3D, t: u.Quantity["time"], /
) -> tuple[AbstractPos3D, u.Quantity["time"]]:
    """Apply the rotation to the coordinates.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    >>> Rz = jnp.asarray([[0, -1, 0], [1, 0,  0], [0, 0, 1]])
    >>> op = cx.ops.GalileanRotation(Rz)

    >>> q = cx.CartesianPos3D.from_([1, 0, 0], "m")
    >>> t = u.Quantity(1, "s")
    >>> newq, newt = op(q, t)
    >>> newq.x
    Quantity['length'](Array(0, dtype=int32), unit='m')

    The time is not affected by the rotation.
    >>> newt
    Quantity['time'](Array(1, dtype=int32, ...), unit='s')

    """
    return self(q), t


@jax.jit
@AbstractOperator.__call__.dispatch
def call(
    self: GalileanRotation, qvec: AbstractPos, pvec: AbstractVel, /, **__: Any
) -> tuple[AbstractPos, AbstractVel]:
    r"""Apply the rotation to the coordinates and velocities.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> R_z = cx.ops.GalileanRotation.from_euler("z", u.Quantity(90, "deg"))

    >>> q = cx.CartesianPos3D.from_([1, 0, 0], "m")
    >>> p = cx.CartesianVel3D.from_([1, 0, 0], "m/s")

    >>> newq, newp = R_z(q, p)
    >>> print(newq, newp, sep="\n")
    <CartesianPos3D (x[m], y[m], z[m])
        [0. 1. 0.]>
    <CartesianVel3D (d_x[m / s], d_y[m / s], d_z[m / s])
        [0. 1. 0.]>

    """
    # Rotate the position.
    newqvec = self(qvec)

    # TODO: figure out how to do this without converting back to arrays.
    # XVel -> CartVel -> Q -> R@Q -> CartVel -> XVel
    pcvec = pvec.vconvert(pvec._cartesian_cls, qvec)  # noqa: SLF001
    p = convert(pcvec.uconvert(ToUnitsOptions.consistent), u.Quantity)
    newp = vec_matmul(self.rotation, p)
    newpcvec = pvec._cartesian_cls.from_(newp)  # noqa: SLF001
    newpvec = newpcvec.vconvert(type(pvec), newqvec)

    return newqvec, newpvec


@AbstractOperator.__call__.dispatch
def call(
    self: GalileanRotation,
    q: u.Quantity["length"],
    p: u.Quantity["speed"],
    /,
    **__: Any,
) -> tuple[u.Quantity["length"], u.Quantity["speed"]]:
    r"""Apply the rotation to the coordinates and velocities.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> R_z = cx.ops.GalileanRotation(jnp.asarray([[0, -1, 0], [1, 0,  0], [0, 0, 1]]))

    >>> q = u.Quantity([1., 0, 0], "m")
    >>> p = u.Quantity([1., 0, 0], "m/s")

    >>> newq, newp = R_z(q, p)
    >>> print(newq, newp, sep="\n")
    Quantity['length'](Array([0., 1., 0.], dtype=float32), unit='m')
    Quantity['speed'](Array([0., 1., 0.], dtype=float32), unit='m / s')

    """
    newq = self(q)
    newp = vec_matmul(self.rotation, p)
    return newq, newp


# ============================================================================
# Simplification


@dispatch
def simplify_op(op: GalileanRotation, /, **kwargs: Any) -> AbstractOperator:
    """Simplify the Galilean rotation operator.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    An operator with a non-identity rotation matrix is not simplified:

    >>> Rz = jnp.asarray([[0, -1, 0], [1, 0,  0], [0, 0, 1]])
    >>> op = cx.ops.GalileanRotation(Rz)
    >>> cx.ops.simplify_op(op)
    GalileanRotation(rotation=i32[3,3])

    An operator with an identity rotation matrix is simplified:

    >>> op = cx.ops.GalileanRotation(jnp.eye(3))
    >>> cx.ops.simplify_op(op)
    Identity()

    """
    if jnp.allclose(op.rotation, jnp.eye(3), **kwargs):
        return Identity()
    return op


@GalileanRotation.__matmul__.dispatch  # type: ignore[misc]
def matmul(self: GalileanRotation, other: GalileanRotation) -> GalileanRotation:
    """Combine two Galilean rotations.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    Two rotations can be combined:

    >>> theta1 = u.Quantity(45, "deg")
    >>> Rz1 = jnp.asarray([[jnp.cos(theta1), -jnp.sin(theta1), 0],
    ...                   [jnp.sin(theta1), jnp.cos(theta1),  0],
    ...                   [0,             0,              1]])
    >>> op1 = cx.ops.GalileanRotation(Rz1)

    >>> theta2 = u.Quantity(90, "deg")
    >>> Rz2 = jnp.asarray([[jnp.cos(theta2), -jnp.sin(theta2), 0],
    ...                   [jnp.sin(theta2), jnp.cos(theta2),  0],
    ...                   [0,             0,              1]])
    >>> op2 = cx.ops.GalileanRotation(Rz2)

    >>> op3 = op1 @ op2
    >>> op3
    GalileanRotation(rotation=f32[3,3])

    >>> jnp.allclose(op3.rotation, op1.rotation @ op2.rotation)
    Array(True, dtype=bool)

    """
    return GalileanRotation(rotation=self.rotation @ other.rotation)


@dispatch
def simplify_op(op1: GalileanRotation, op2: GalileanRotation) -> GalileanRotation:
    """Combine two Galilean rotations.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    Two rotations can be combined:

    >>> theta1 = u.Quantity(45, "deg")
    >>> Rz1 = jnp.asarray([[jnp.cos(theta1), -jnp.sin(theta1), 0],
    ...                   [jnp.sin(theta1), jnp.cos(theta1),  0],
    ...                   [0,             0,              1]])
    >>> op1 = cx.ops.GalileanRotation(Rz1)

    >>> theta2 = u.Quantity(60, "deg")
    >>> Rz2 = jnp.asarray([[jnp.cos(theta2), -jnp.sin(theta2), 0],
    ...                   [jnp.sin(theta2), jnp.cos(theta2),  0],
    ...                   [0,             0,              1]])
    >>> op2 = cx.ops.GalileanRotation(Rz2)

    >>> op3 = cx.ops.simplify_op(op1, op2)
    >>> op3
    GalileanRotation(rotation=f32[3,3])

    >>> jnp.allclose(op3.rotation, op1.rotation @ op2.rotation)
    Array(True, dtype=bool)

    """
    return GalileanRotation(rotation=op1.rotation @ op2.rotation)



"""Built-in vector classes."""

__all__ = [
    "CartesianGeneric3D",
]

from typing import TypeVar, final

import equinox as eqx
import jax

import quaxed.numpy as jnp
from unxt.quantity import Quantity

import coordinax._src.typing as ct
from coordinax._src.vectors.base import AbstractVector
from coordinax._src.vectors.mixins import AvalMixin

VT = TypeVar("VT", bound="CartesianGeneric3D")


@final
class CartesianGeneric3D(AvalMixin, AbstractVector):
    """Generic 3D Cartesian coordinates.

    The fields of this class are not restricted to any specific dimensions.

    Examples
    --------
    >>> import coordinax as cx
    >>> vec = cx.vecs.CartesianGeneric3D.from_([1, 2, 3], "kg m /s")
    >>> print(vec)
    <CartesianGeneric3D (x[kg m / s], y[kg m / s], z[kg m / s])
        [1 2 3]>

    """

    x: ct.BatchableScalarQ = eqx.field(converter=Quantity.from_)

    y: ct.BatchableScalarQ = eqx.field(converter=Quantity.from_)

    z: ct.BatchableScalarQ = eqx.field(converter=Quantity.from_)

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianGeneric3D._dimensionality()
        3

        """
        return 3

    # -----------------------------------------------------
    # Unary operations

    def __neg__(self) -> "CartesianGeneric3D":
        """Negate the `coordinax.CartesianGeneric3D`.

        Examples
        --------
        >>> import coordinax as cx
        >>> q = cx.vecs.CartesianGeneric3D.from_([1, 2, 3], "km")
        >>> print(-q)
        <CartesianGeneric3D (x[km], y[km], z[km])
        [-1 -2 -3]>

        """
        return jax.tree.map(jnp.negative, self)



"""Built-in vector classes."""

__all__ = ["ProlateSpheroidalAcc", "ProlateSpheroidalPos", "ProlateSpheroidalVel"]

from dataclasses import KW_ONLY
from typing import final

import equinox as eqx
from jaxtyping import Shaped

import quaxed.numpy as jnp
import unxt as u
from dataclassish.converters import Unless

import coordinax._src.typing as ct
from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D
from coordinax._src.angles import Angle, BatchableAngleQ
from coordinax._src.utils import classproperty
from coordinax._src.vectors.base import VectorAttribute
from coordinax._src.vectors.checks import (
    check_greater_than_equal,
    check_less_than_equal,
    check_non_negative_non_zero,
)
from coordinax._src.vectors.converters import converter_azimuth_to_range


@final
class ProlateSpheroidalPos(AbstractPos3D):
    """Prolate spheroidal coordinates as defined by Dejonghe & de Zeeuw 1988.

    Note that valid coordinates have:
    - mu >= Delta^2
    - |nu| <= Delta^2
    - Delta > 0

    Parameters
    ----------
    mu : Quantity["area"]
        The spheroidal mu coordinate. This is called `lambda` by Dejonghe & de Zeeuw.
    nu : Quantity["area"]
        The spheroidal nu coordinate.
    phi : `coordinax.angle.Angle`
        Azimuthal angle [0, 360) [deg] where 0 is the x-axis.
    Delta : Quantity["length"]
        The focal length of the coordinate system. Must be > 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.vecs.ProlateSpheroidalPos(
    ...     mu=u.Quantity(3.0, "km2"),
    ...     nu=u.Quantity(0.5, "km2"),
    ...     phi=u.Quantity(0.25, "rad"),
    ...     Delta=u.Quantity(1.5, "km"),
    ... )
    >>> vec
    ProlateSpheroidalPos(
      mu=Quantity[PhysicalType('area')](value=...f32[], unit=Unit("km2")),
      nu=Quantity[PhysicalType('area')](value=...f32[], unit=Unit("km2")),
      phi=Angle(value=...f32[], unit=Unit("rad")),
      Delta=Quantity[PhysicalType('length')](value=...f32[], unit=Unit("km"))
    )

    This fails with a zero or negative Delta:

    >>> try: vec = cx.vecs.ProlateSpheroidalPos(
    ...     mu=u.Quantity(3.0, "km2"),
    ...     nu=u.Quantity(0.5, "km2"),
    ...     phi=u.Quantity(0.25, "rad"),
    ...     Delta=u.Quantity(0.0, "km"),
    ... )
    ... except Exception as e: pass

    Or with invalid mu and nu:

    >>> try: vec = cx.vecs.ProlateSpheroidalPos(
    ...     mu=u.Quantity(0.5, "km2"),
    ...     nu=u.Quantity(0.5, "km2"),
    ...     phi=u.Quantity(0.25, "rad"),
    ...     Delta=u.Quantity(1.5, "km"),
    ... )
    ... except Exception as e: pass

    """

    mu: ct.BatchableArea = eqx.field(converter=u.Quantity["area"].from_)
    r"""Spheroidal mu coordinate :math:`\mu \in [0,+\infty)` (called :math:`\lambda` in
     some Galactic dynamics contexts)."""

    nu: ct.BatchableArea = eqx.field(converter=u.Quantity["area"].from_)
    r"""Spheroidal nu coordinate :math:`\lambda \in [-\infty,+\infty)`."""

    phi: BatchableAngleQ = eqx.field(
        converter=Unless(Angle, lambda x: converter_azimuth_to_range(Angle.from_(x)))
    )
    r"""Azimuthal angle, generally :math:`\phi \in [0,360)`."""

    _: KW_ONLY
    Delta: Shaped[u.Quantity["length"], ""] = VectorAttribute()
    """Focal length of the coordinate system."""

    def __check_init__(self) -> None:
        """Check the validity of the initialization."""
        check_non_negative_non_zero(self.Delta, name="Delta")
        check_greater_than_equal(
            self.mu, self.Delta**2, name="mu", comparison_name="Delta^2"
        )
        check_less_than_equal(
            jnp.abs(self.nu), self.Delta**2, name="nu", comparison_name="Delta^2"
        )

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["ProlateSpheroidalVel"]:
        return ProlateSpheroidalVel


@final
class ProlateSpheroidalVel(AbstractVel3D):
    """Prolate spheroidal differential representation."""

    d_mu: ct.BatchableDiffusivity = eqx.field(converter=u.Quantity["diffusivity"].from_)
    r"""Prolate spheroidal mu speed :math:`d\mu/dt \in [-\infty, \infty]."""

    d_nu: ct.BatchableDiffusivity = eqx.field(converter=u.Quantity["diffusivity"].from_)
    r"""Prolate spheroidal nu speed :math:`d\nu/dt \in [-\infty, \infty]."""

    d_phi: ct.BatchableAngularSpeed = eqx.field(
        converter=u.Quantity["angular speed"].from_
    )
    r"""Azimuthal speed :math:`d\phi/dt \in [-\infty, \infty]."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[ProlateSpheroidalPos]:
        return ProlateSpheroidalPos

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["ProlateSpheroidalAcc"]:
        return ProlateSpheroidalAcc


@final
class ProlateSpheroidalAcc(AbstractAcc3D):
    """Prolate spheroidal acceleration representation."""

    d2_mu: ct.BatchableSpecificEnergy = eqx.field(
        converter=u.Quantity["specific energy"].from_
    )
    r"""Prolate spheroidal mu acceleration :math:`d^2\mu/dt^2 \in [-\infty, \infty]."""

    d2_nu: ct.BatchableSpecificEnergy = eqx.field(
        converter=u.Quantity["specific energy"].from_
    )
    r"""Prolate spheroidal nu acceleration :math:`d^2\nu/dt^2 \in [-\infty, \infty]."""

    d2_phi: ct.BatchableAngularAcc = eqx.field(
        converter=u.Quantity["angular acceleration"].from_
    )
    r"""Azimuthal acceleration :math:`d^2\phi/dt^2 \in [-\infty, \infty]."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[ProlateSpheroidalVel]:
        return ProlateSpheroidalVel



"""Distance constructors."""

__all__: list[str] = []

from typing import Any

import jax.numpy as jnp

import quaxed.numpy as jnp
import unxt as u
from unxt.quantity import AbstractQuantity, UncheckedQuantity

from .measures import Distance, DistanceModulus, Parallax

parallax_base_length = u.Quantity(1, "AU")
distance_modulus_base_distance = u.Quantity(10, "pc")

# -------------------------------------------------------------------
# To Distance


@AbstractQuantity.from_.dispatch
def from_(cls: type[Distance], d: Distance) -> Distance:
    """Construct a `Distance` from a `Distance`.

    Examples
    --------
    >>> from coordinax.distance import Distance

    >>> d = Distance(1, "pc")
    >>> Distance.from_(d) is d
    True

    """
    return d


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[Distance], p: Parallax | u.Quantity["angle"], /, **kwargs: Any
) -> Distance:
    """Construct a `coordinax.Distance` from an angle through the parallax.

    Examples
    --------
    >>> import unxt as u
    >>> from coordinax.distance import Distance, Parallax

    >>> Distance.from_(Parallax(1, "mas")).uconvert("kpc")
    Distance(Array(1., dtype=float32, ...), unit='kpc')

    >>> Distance.from_(u.Quantity(1, "mas")).uconvert("kpc")
    Distance(Array(1., dtype=float32, ...), unit='kpc')

    """
    d = parallax_base_length / jnp.tan(p)  # [AU]
    unit = u.unit_of(d)
    return cls(jnp.asarray(d.ustrip(unit), **kwargs), unit)


@AbstractQuantity.from_.dispatch  # type: ignore[no-redef]
def from_(
    cls: type[Distance], dm: DistanceModulus | u.Quantity["mag"], /, **kwargs: Any
) -> Distance:
    """Construct a `Distance` from a mag through the dist mod.

    Examples
    --------
    >>> import unxt as u
    >>> from coordinax.distance import Distance, DistanceModulus

    >>> Distance.from_(DistanceModulus(10, "mag")).uconvert("pc")
    Distance(Array(1000., dtype=float32, ...), unit='pc')

    >>> Distance.from_(u.Quantity(10, "mag")).uconvert("pc")
    Distance(Array(1000., dtype=float32, ...), unit='pc')

    """
    d = 10 ** (1 + dm.ustrip("mag") / 5)
    return cls(jnp.asarray(d, **kwargs), "pc")


# -------------------------------------------------------------------
# To DistanceModulus


@AbstractQuantity.from_.dispatch
def from_(cls: type[DistanceModulus], dm: DistanceModulus) -> DistanceModulus:
    """Construct a `DistanceModulus` from a `DistanceModulus`.

    Examples
    --------
    >>> from coordinax.distance import DistanceModulus

    >>> dm = DistanceModulus(1, "mag")
    >>> DistanceModulus.from_(dm) is dm
    True

    """
    return dm


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[DistanceModulus], d: Distance | u.Quantity["length"], /, **kwargs: Any
) -> DistanceModulus:
    """Construct a `DistanceModulus` from a distance.

    Examples
    --------
    >>> import unxt as u
    >>> from coordinax.distance import Distance, DistanceModulus

    >>> DistanceModulus.from_(Distance(1, "pc"))
    DistanceModulus(Array(-5., dtype=float32), unit='mag')

    >>> DistanceModulus.from_(u.Quantity(1, "pc"))
    DistanceModulus(Array(-5., dtype=float32), unit='mag')

    """
    dm = 5 * jnp.log10(d.ustrip("pc")) - 5
    return cls(jnp.asarray(dm, **kwargs), "mag")


@AbstractQuantity.from_.dispatch  # type: ignore[no-redef]
def from_(
    cls: type[DistanceModulus], p: Parallax | u.Quantity["angle"], /, **kwargs: Any
) -> DistanceModulus:
    """Construct a `DistanceModulus` from a parallax.

    Examples
    --------
    >>> import unxt as u
    >>> from coordinax.distance import DistanceModulus, Parallax

    >>> DistanceModulus.from_(Parallax(1, "mas"))
    DistanceModulus(Array(10., dtype=float32), unit='mag')

    >>> DistanceModulus.from_(u.Quantity(1, "mas"))
    DistanceModulus(Array(10., dtype=float32), unit='mag')

    """
    d = parallax_base_length / jnp.tan(p)  # [AU]
    dm = 5 * jnp.log10(d.ustrip("pc")) - 5
    return cls(jnp.asarray(dm, **kwargs), "mag")


# -------------------------------------------------------------------
# To Parallax


@AbstractQuantity.from_.dispatch
def from_(cls: type[Parallax], p: Parallax) -> Parallax:
    """Construct a `Parallax` from a `Parallax`.

    Examples
    --------
    >>> from coordinax.distance import Parallax

    >>> p = Parallax(1, "mas")
    >>> Parallax.from_(p) is p
    True

    """
    return p


@AbstractQuantity.from_.dispatch  # type: ignore[no-redef]
def from_(
    cls: type[Parallax], d: Distance | u.Quantity["length"], /, **kwargs: Any
) -> Parallax:
    """Construct a `Parallax` from a distance.

    Examples
    --------
    >>> import unxt as u
    >>> from coordinax.distance import Parallax, Distance

    >>> Parallax.from_(Distance(10, "pc")).uconvert("mas").round(2)
    Parallax(Array(100., dtype=float32, ...), unit='mas')

    >>> Parallax.from_(u.Quantity(10, "pc")).uconvert("mas").round(2)
    Parallax(Array(100., dtype=float32, ...), unit='mas')

    """
    p = jnp.atan2(parallax_base_length, d)
    return cls(jnp.asarray(p.value, **kwargs), p.unit)


@AbstractQuantity.from_.dispatch  # type: ignore[no-redef]
def from_(
    cls: type[Parallax], dm: DistanceModulus | u.Quantity["mag"], /, **kwargs: Any
) -> Parallax:
    """Construct a `Parallax` from a distance.

    Examples
    --------
    >>> import unxt as u
    >>> from coordinax.distance import Parallax, DistanceModulus

    >>> Parallax.from_(DistanceModulus(23, "mag")).uconvert("mas").round(2)
    Parallax(Array(0., dtype=float32, ...), unit='mas')

    >>> Parallax.from_(u.Quantity(23, "mag")).uconvert("mas").round(2)
    Parallax(Array(0., dtype=float32, ...), unit='mas')

    """
    d = UncheckedQuantity(10 ** (1 + dm.ustrip("mag") / 5), "pc")
    p = jnp.atan2(parallax_base_length, d)
    unit = u.unit_of(p)
    return cls(jnp.asarray(p.ustrip(unit), **kwargs), unit)



# ruff: noqa: ERA001
"""Galilean coordinate transformations."""

__all__ = ["GalileanTranslation"]


from typing import Any, Literal, final

import equinox as eqx
from plum import convert, dispatch

import quaxed.numpy as jnp
import unxt as u

from .base import AbstractGalileanOperator
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.operators.identity import Identity
from coordinax._src.vectors.d3 import AbstractPos3D
from coordinax._src.vectors.d4 import FourVector


@final
class GalileanTranslation(AbstractGalileanOperator):
    r"""Operator for spatio-temporal translations.

    The coordinate transform is given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t+s, \mathbf{x} + \mathbf {a})

    where :math:`a \in R^3` and :math:`s \in R`.  Therefore for a potential
    :math:`\Phi(t,\mathbf{x})` in the translated frame the potential is given by
    the subtraction of the translation.

    Parameters
    ----------
    translation : :class:`vector.FourVector`
        The translation vector [T, Q].  This parameters uses
        :meth:`vector.FourVector.from_` to enable a variety of more
        convenient input types. See :class:`vector.FourVector` for details.

    Examples
    --------
    We start with the required imports:

    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    We can then create a translation operator:

    >>> op = cx.ops.GalileanTranslation.from_([1.0, 2.0, 3.0, 4.0], "km")
    >>> op
    GalileanTranslation(FourVector(
        t=Quantity[PhysicalType('time')](value=f32[], unit=Unit("s")),
        q=CartesianPos3D( ... ) ))

    Note that the translation is a :class:`vector.FourVector`, which was
    constructed from a 1D array, using :meth:`vector.FourVector.from_`. We
    can also construct it directly, which allows for other vector types.

    >>> qshift = cx.SphericalPos(r=u.Quantity(1.0, "km"),
    ...                          theta=u.Quantity(jnp.pi/2, "rad"),
    ...                          phi=u.Quantity(0, "rad"))
    >>> shift = cx.FourVector(u.Quantity(1.0, "Gyr"), qshift)
    >>> op = cx.ops.GalileanTranslation(shift)
    >>> op
    GalileanTranslation(FourVector(
        t=Quantity[PhysicalType('time')](value=...f32[], unit=Unit("Gyr")),
        q=SphericalPos( ... ) ))

    Translation operators can be applied to :class:`vector.FourVector`:

    >>> w = cx.FourVector.from_([0, 0, 0, 0], "km")
    >>> op(w)
    FourVector(
      t=Quantity[PhysicalType('time')](value=...f32[], unit=Unit("s")),
      q=CartesianPos3D( ... )
    )

    Also to :class:`vector.AbstractPos3D` and :class:`unxt.Quantity`:

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "km")
    >>> t = u.Quantity(0, "Gyr")
    >>> newq, newt = op(q, t)
    >>> newq.x
    Quantity['length'](Array(1., dtype=float32, ...), unit='km')
    >>> newt
    Quantity['time'](Array(1., dtype=float32, ...), unit='Gyr')

    """

    translation: FourVector = eqx.field(converter=FourVector.from_)
    """The temporal + spatial translation.

    The translation vector [T, Q].  This parameters uses
    :meth:`vector.FourVector.from_` to enable a variety of more convenient
    input types. See :class:`vector.FourVector` for details.
    """

    # -------------------------------------------

    @property
    def is_inertial(self) -> Literal[True]:
        """Galilean translation is an inertial-frame preserving transformation.

        Examples
        --------
        >>> import coordinax as cx

        >>> shift = cx.FourVector.from_([0, 1, 1, 1], "km")
        >>> op = cx.ops.GalileanTranslation(shift)

        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "GalileanTranslation":
        """The inverse of the operator.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> qshift = cx.CartesianPos3D.from_([1, 1, 1], "km")
        >>> shift = FourVector(u.Quantity(1, "Gyr"), qshift)
        >>> op = cx.ops.GalileanTranslation(shift)

        >>> op.inverse
        GalileanTranslation(FourVector( ... ))

        >>> op.inverse.translation.q.x
        Quantity['length'](Array(-1, dtype=int32), unit='km')

        """
        return GalileanTranslation(-self.translation)

    # -------------------------------------------

    @AbstractOperator.__call__.dispatch
    def __call__(
        self: "GalileanTranslation", x: FourVector, /, **__: Any
    ) -> FourVector:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx
        >>> import coordinax.ops as cxo

        Explicitly construct the translation operator:

        >>> qshift = cx.CartesianPos3D.from_([1, 1, 1], "km")
        >>> shift = FourVector(u.Quantity(1, "Gyr"), qshift)
        >>> op = cx.ops.GalileanTranslation(shift)

        Construct a vector to translate, using the convenience from_ (the
        0th component is :math:`c * t`, the rest are spatial components):

        >>> w = cx.FourVector.from_([0, 1, 2, 3], "km")
        >>> w.t
        Quantity['time'](Array(0., dtype=float32, ...), unit='s')

        Apply the translation operator:

        >>> new = op(w)
        >>> new.x
        Quantity['length'](Array(2, dtype=int32), unit='km')

        >>> new.t.uconvert("Gyr")
        Quantity['time'](Array(1., dtype=float32, ...), unit='Gyr')

        """
        return x + self.translation

    @AbstractOperator.__call__.dispatch
    def __call__(
        self: "GalileanTranslation",
        x: AbstractPos3D,
        t: u.Quantity["time"],
        /,
        **__: Any,
    ) -> tuple[AbstractPos3D, u.Quantity["time"]]:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx
        >>> import coordinax.ops as cxo

        Explicitly construct the translation operator:

        >>> qshift = cx.CartesianPos3D.from_([1, 1, 1], "km")
        >>> tshift = u.Quantity(1, "Gyr")
        >>> shift = cx.FourVector(tshift, qshift)
        >>> op = cx.ops.GalileanTranslation(shift)

        Construct a vector to translate

        >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
        >>> t = u.Quantity(1, "Gyr")
        >>> newq, newt = op(q, t)

        >>> newq.x
        Quantity['length'](Array(2, dtype=int32), unit='km')

        >>> newt
        Quantity['time'](Array(2, dtype=int32, ...), unit='Gyr')

        """
        return (x + self.translation.q, t + self.translation.t)

    # -------------------------------------------
    # Python special methods

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.translation!r})"


@dispatch
def simplify_op(
    op: GalileanTranslation, /, **kwargs: Any
) -> GalileanTranslation | Identity:
    """Simplify a Galilean translation operator.

    Examples
    --------
    >>> import coordinax as cx

    An operator with real effect cannot be simplified:

    >>> op = cx.ops.GalileanTranslation.from_([3e8, 1, 0, 0], "m")
    >>> cx.ops.simplify_op(op)
    GalileanTranslation(FourVector( ... ))

    An operator with no effect can be simplified:

    >>> op = cx.ops.GalileanTranslation.from_([0, 0, 0, 0], "m")
    >>> cx.ops.simplify_op(op)
    Identity()

    """
    # Check if the translation is zero.
    if jnp.allclose(
        convert(op.translation, u.Quantity).value, jnp.zeros((4,)), **kwargs
    ):
        return Identity()
    # TODO: Check if the translation is purely spatial.

    return op


# TODO: show op3.translation = op1.translation + op2.translation
@dispatch
def simplify_op(
    op1: GalileanTranslation, op2: GalileanTranslation, /
) -> GalileanTranslation:
    """Combine two translations into a single translation.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> qshift = cx.CartesianPos3D.from_([1, 0, 0], "km")
    >>> tshift = u.Quantity(1, "Gyr")
    >>> op1 = cx.ops.GalileanTranslation(FourVector(tshift, qshift))

    >>> qshift = cx.CartesianPos3D.from_([0, 1, 0], "km")
    >>> tshift = u.Quantity(1, "Gyr")
    >>> op2 = cx.ops.GalileanTranslation(FourVector(tshift, qshift))

    >>> op3 = cx.ops.simplify_op(op1, op2)
    >>> op3
    GalileanTranslation(FourVector( ... ))

    """
    return GalileanTranslation(op1.translation + op2.translation)



"""Transformations on Frames."""

__all__: list[str] = ["TransformedReferenceFrame"]

from typing import final

from plum import dispatch

from .base import AbstractReferenceFrame
from coordinax._src.operators.base import AbstractOperator


@final
class TransformedReferenceFrame(AbstractReferenceFrame):
    r"""Transformations relative to a base reference frame.

    This class represents a reference frame that is defined relative to a base
    reference frame. The transformation from the base reference frame to this
    reference frame is stored as an `coordinax.operators.AbstractOperator`.

    :::{warning}

    The transformation operator ``ops`` is the transformation of the
    **frames**, not the coordinates. This is a passive (intrinsic) versus active
    (extrinsic) transformation. A frame transformation (passive) is applied to
    the coordinates by the inverse of the operator.

    :::

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx
    >>> import coordinax.frames as cxf

    >>> R = cx.ops.GalileanRotation([[0., -1, 0], [1, 0, 0], [0, 0, 1]])
    >>> frame = cxf.TransformedReferenceFrame(cxf.ICRS(), R)
    >>> frame
    TransformedReferenceFrame(
      base_frame=ICRS(), xop=GalileanRotation(rotation=f32[3,3])
    )

    Let's transform a position from the base frame to the transformed frame:

    >>> op = cxf.frame_transform_op(cxf.ICRS(), frame)

    >>> q_icrs = cx.CartesianPos3D.from_([1, 0, 0], "kpc")
    >>> q_frame = op(q_icrs)
    >>> print(q_frame)
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [ 0. -1.  0.]>

    >>> op.inverse(q_frame) == q_icrs
    Array(True, dtype=bool)

    This can also transform a velocity:

    >>> v_icrs = cx.CartesianVel3D.from_([1, 0, 0], "km/s")
    >>> q_frame, v_frame = op(q_icrs, v_icrs)
    >>> print(q_frame, v_frame, sep="\n")
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [ 0. -1.  0.]>
    <CartesianVel3D (d_x[km / s], d_y[km / s], d_z[km / s])
        [ 0. -1.  0.]>

    >>> op.inverse(q_frame, v_frame) == (q_icrs, v_icrs)
    True

    """

    #: The base reference frame.
    base_frame: AbstractReferenceFrame

    #: The transformation from the base frame to this frame.
    #: This is a passive transformation, describing the transformation of the
    #: frame, not the coordinates. To transform the coordinates, apply the
    #: inverse of this operator.
    xop: AbstractOperator


@dispatch
def frame_transform_op(
    from_frame: AbstractReferenceFrame, to_frame: TransformedReferenceFrame
) -> AbstractOperator:
    """Return a frame transform operator to a transformed frame.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx
    >>> import coordinax.frames as cxf

    >>> R = cx.ops.GalileanRotation([[0., -1, 0], [1, 0, 0], [0, 0, 1]])
    >>> frame = cxf.TransformedReferenceFrame(cxf.ICRS(), R)
    >>> frame
    TransformedReferenceFrame(
      base_frame=ICRS(), xop=GalileanRotation(rotation=f32[3,3])
    )

    Let's transform a position from the base frame to the transformed frame:

    >>> op = cxf.frame_transform_op(cxf.ICRS(), frame)

    >>> q_icrs = cx.CartesianPos3D.from_([1, 0, 0], "kpc")
    >>> q_frame = op(q_icrs)
    >>> print(q_frame)
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [ 0. -1.  0.]>

    """
    return frame_transform_op(from_frame, to_frame.base_frame) | to_frame.xop.inverse


@dispatch
def frame_transform_op(
    from_frame: TransformedReferenceFrame, to_frame: AbstractReferenceFrame
) -> AbstractOperator:
    """Return a frame transform operator from a transformed frame.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx
    >>> import coordinax.frames as cxf

    >>> R = cx.ops.GalileanRotation([[0., -1, 0], [1, 0, 0], [0, 0, 1]])
    >>> frame = cxf.TransformedReferenceFrame(cxf.ICRS(), R)
    >>> frame
    TransformedReferenceFrame(
      base_frame=ICRS(), xop=GalileanRotation(rotation=f32[3,3])
    )

    Let's transform a position from the base frame to the transformed frame:

    >>> op = cxf.frame_transform_op(frame, cxf.ICRS())

    >>> q_icrs = cx.CartesianPos3D.from_([0, -1, 0], "kpc")
    >>> q_frame = op(q_icrs)
    >>> print(q_frame)
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [1. 0. 0.]>

    """
    return from_frame.xop | frame_transform_op(from_frame.base_frame, to_frame)


@dispatch(precedence=1)
def frame_transform_op(
    from_frame: TransformedReferenceFrame, to_frame: TransformedReferenceFrame
) -> AbstractOperator:
    """Return a frame transform operator between two transformed frames.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx
    >>> import coordinax.frames as cxf

    >>> R = cx.ops.GalileanRotation([[0., -1, 0], [1, 0, 0], [0, 0, 1]])
    >>> frame1 = cxf.TransformedReferenceFrame(cxf.ICRS(), R)

    >>> shift = cx.ops.GalileanSpatialTranslation.from_([1, 0, 0], "kpc")
    >>> frame2 = cxf.TransformedReferenceFrame(frame1, shift)

    >>> op1to2 = cxf.frame_transform_op(frame1, frame2)

    >>> q_frame1 = cx.CartesianPos3D.from_([0, -1, 0], "kpc")
    >>> q_frame2 = op1to2(q_icrs)

    """
    return (
        from_frame.xop
        | frame_transform_op(from_frame.base_frame, to_frame.base_frame)
        | to_frame.xop.inverse
    )



"""Abstract Velocity."""

__all__ = ["AbstractVel", "VELOCITY_CLASSES"]

from .core import VELOCITY_CLASSES, AbstractVel

# Register by import
# isort: split
from . import register_primitives  # noqa: F401



"""Base classes for operators on coordinates and potentials."""

__all__ = ["AbstractCompositeOperator"]

from collections.abc import Iterator
from dataclasses import replace
from typing import TYPE_CHECKING, Any, Protocol, overload, runtime_checkable

from dataclassish import DataclassInstance

from .base import AbstractOperator
from coordinax._src.vectors.base import AbstractVector

if TYPE_CHECKING:
    from typing import Self


@runtime_checkable
class HasOperatorsAttr(DataclassInstance, Protocol):  # type: ignore[misc]
    """Protocol for classes with an `operators` attribute."""

    operators: tuple[AbstractOperator, ...]


class AbstractCompositeOperator(AbstractOperator):
    """Abstract Composite Operator.

    This is the base class for all composite operations.

    See Also
    --------
    :class:`coordinax.ops.Pipe`
    :class:`coordinax.ops.GalileanOperator`

    """

    # ===========================================
    # Operator

    @property
    def is_inertial(self: HasOperatorsAttr) -> bool:
        """Whether the operations maintain an inertial reference frame."""
        return all(op.is_inertial for op in self.operators)

    @property
    def inverse(self) -> "Pipe":
        """The inverse of the operator.

        This is the sequence of the inverse of each operator in reverse order.

        Examples
        --------
        >>> import coordinax as cx

        >>> shift = cx.ops.GalileanSpatialTranslation.from_([1, 2, 3], "km")
        >>> boost = cx.ops.GalileanBoost.from_([1, 2, 3], "km/s")
        >>> pipe = cx.ops.Pipe((shift, boost))
        >>> pipe.inverse
        Pipe(( GalileanBoost(...), GalileanSpatialTranslation(...) ))

        """
        from .pipe import Pipe

        return Pipe(tuple(op.inverse for op in reversed(self.operators)))

    @AbstractOperator.__call__.dispatch(precedence=1)  # type: ignore[attr-defined, misc]
    def __call__(
        self: "AbstractCompositeOperator", *args: object, **kwargs: Any
    ) -> tuple[object, ...]:
        """Apply the operators to the coordinates.

        This is the default implementation, which applies the operators in
        sequence, passing along the arguments.

        Examples
        --------
        >>> import coordinax as cx

        >>> op = cx.ops.Pipe((cx.ops.Identity(), cx.ops.Identity()))
        >>> op(1, 2, 3)
        (1, 2, 3)

        """
        for op in self.operators:
            args = op(*args, **kwargs)
        return args

    # ===========================================
    # Pipe

    @overload
    def __getitem__(self, key: int) -> AbstractOperator: ...

    @overload
    def __getitem__(self, key: slice) -> "Self": ...

    def __getitem__(self, key: int | slice) -> "AbstractOperator | Self":
        ops = self.operators[key]
        if isinstance(ops, AbstractOperator):
            return ops
        return replace(self, operators=ops)

    def __iter__(self: HasOperatorsAttr) -> Iterator[AbstractOperator]:
        return iter(self.operators)


# ======================================================================
# Call dispatches


@AbstractOperator.__call__.dispatch(precedence=1)  # type: ignore[attr-defined, misc]
def call(
    self: AbstractCompositeOperator, x: AbstractVector, /, **kwargs: Any
) -> AbstractVector:
    """Apply the operator to the coordinates.

    This is the default implementation, which applies the operators in
    sequence.

    Examples
    --------
    >>> import coordinax as cx

    >>> op = cx.ops.Pipe((cx.ops.Identity(), cx.ops.Identity()))
    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> op(vec)
    CartesianPos3D( ... )

    """
    # TODO: with lax.for_i
    for op in self.operators:
        x = op(x, **kwargs)
    return x



"""Astronomy reference frames."""
# ruff:noqa: N806

__all__: list[str] = []


from typing import TypeAlias

from jaxtyping import Array, Shaped
from plum import dispatch

import quaxed.numpy as jnp
import unxt as u

from .galactocentric import Galactocentric
from .icrs import ICRS
from coordinax._src.angles import Angle
from coordinax._src.distances import Distance
from coordinax._src.operators import (
    GalileanRotation,
    GalileanSpatialTranslation,
    Identity,
    Pipe,
    VelocityBoost,
    simplify_op,
)

ScalarAngle: TypeAlias = Shaped[u.Quantity["angle"] | Angle, ""]
RotationMatrix: TypeAlias = Shaped[Array, "3 3"]
LengthVector: TypeAlias = Shaped[u.Quantity["length"], "3"] | Shaped[Distance, "3"]
VelocityVector: TypeAlias = Shaped[u.Quantity["speed"], "3"]


@dispatch
def frame_transform_op(from_frame: ICRS, to_frame: ICRS, /) -> Identity:
    """Return an identity operator for the ICRS->ICRS transformation.

    Examples
    --------
    >>> import coordinax.frames as cxf
    >>> icrs_frame = cxf.ICRS()
    >>> frame_op = cxf.frame_transform_op(icrs_frame, icrs_frame)
    >>> frame_op
    Identity()

    """
    return Identity()


# ---------------------------------------------------------------


@dispatch
def frame_transform_op(from_frame: Galactocentric, to_frame: Galactocentric, /) -> Pipe:
    """Return a sequence of operators for the Galactocentric frame self transformation.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.frames as cxf

    >>> gcf_frame = cxf.Galactocentric()
    >>> frame_op = cxf.frame_transform_op(gcf_frame, gcf_frame)
    >>> frame_op
    Pipe((Identity(),))

    >>> gcf_frame2 = cxf.Galactocentric(roll=u.Quantity(10, "deg"))
    >>> frame_op2 = cxf.frame_transform_op(gcf_frame, gcf_frame2)
    >>> frame_op2
    Pipe((
        VelocityBoost(CartesianVel3D( ... )),
        GalileanRotation(rotation=f32[3,3]),
        GalileanSpatialTranslation(CartesianPos3D( ... )),
        GalileanRotation(rotation=f32[3,3]),
        GalileanRotation(rotation=f32[3,3]),
        GalileanSpatialTranslation(CartesianPos3D( ... )),
        GalileanRotation(rotation=f32[3,3]),
        VelocityBoost(CartesianVel3D( ... ))
    ))

    """
    if from_frame == to_frame:
        return Pipe((Identity(),))

    # TODO: not go through ICRS for the self-transformation
    return simplify_op(
        frame_transform_op(from_frame, ICRS()) | frame_transform_op(ICRS(), to_frame)
    )


# ---------------------------------------------------------------


@dispatch
def frame_transform_op(from_frame: ICRS, to_frame: Galactocentric, /) -> Pipe:
    r"""Return an ICRS to Galactocentric frame transformation operator.

    Examples
    --------
    For this example we compare against the Astropy implementation of the
    frame transformation:

    >>> import unxt as u
    >>> import coordinax as cx
    >>> import astropy.coordinates as apyc

    The location of Vega in ICRS coordinates:

    >>> vega = apyc.SkyCoord(
    ...     ra=279.23473479 * u.unit("deg"), dec=38.78368896 * u.unit("deg"),
    ...     distance=25 * u.unit("pc"),
    ...     pm_ra_cosdec=200 * u.unit("mas / yr"), pm_dec=-286 * u.unit("mas / yr"),
    ...     radial_velocity=-13.9 * u.unit("km / s"))
    >>> print(vega)
    <SkyCoord (ICRS): (ra, dec, distance) in (deg, deg, pc)
        (279.23473479, 38.78368896, 25.)
     (pm_ra_cosdec, pm_dec, radial_velocity) in (mas / yr, mas / yr, km / s)
        (200., -286., -13.9)>

    Transforming to a Galactocentric frame:

    >>> apy_gcf = apyc.Galactocentric()
    >>> apy_gcf
    <Galactocentric Frame (galcen_coord=<ICRS Coordinate: (ra, dec) in deg
        (266.4051, -28.936175)>, galcen_distance=8.122 kpc, galcen_v_sun=(12.9, 245.6, 7.78) km / s, z_sun=20.8 pc, roll=0.0 deg)>

    >>> vega.transform_to(apy_gcf)
    <SkyCoord (Galactocentric: ...): (x, y, z) in pc
        (-8112.89970167, 21.79911216, 29.01384942)
     (v_x, v_y, v_z) in km / s
        (34.06711868, 234.61647066, -28.75976702)>

    Now we can use `coordinax` to perform the same transformation!

    Converting the Astropy objects to `coordinax`:

    >>> vega_q = cx.vecs.LonLatSphericalPos.from_(vega.icrs.data)
    >>> vega_p = cx.vecs.LonCosLatSphericalVel.from_(vega.icrs.data.differentials["s"])

    >>> icrs_frame = cx.frames.ICRS()
    >>> gcf_frame = cx.frames.Galactocentric.from_(apy_gcf)

    Define the transformation operator:

    >>> frame_op = cx.frames.frame_transform_op(icrs_frame, gcf_frame)
    >>> frame_op
    Pipe((
        GalileanRotation(rotation=f32[3,3]),
        GalileanSpatialTranslation(CartesianPos3D( ... )),
        GalileanRotation(rotation=f32[3,3]),
        VelocityBoost(CartesianVel3D( ... ))
    ))

    Apply the transformation:

    >>> vega_gcf_q, vega_gcf_p = frame_op(vega_q, vega_p)
    >>> vega_gcf_q = vega_gcf_q.vconvert(cx.vecs.CartesianPos3D)
    >>> vega_gcf_p = vega_gcf_p.vconvert(cx.vecs.CartesianVel3D, vega_gcf_q)
    >>> print(vega_gcf_q)
    <CartesianPos3D (x[pc], y[pc], z[pc])
        [-8112.898    21.799    29.01...]>
    >>> print(vega_gcf_p.uconvert({u.dimension("speed"): "km/s"}))
    <CartesianVel3D (d_x[km / s], d_y[km / s], d_z[km / s])
        [ 34.067 234.616 -28.76 ]>

    It matches!

    Let's do it again for a few different input types:

    >>> q = u.Quantity([0, 0, 0], "pc")
    >>> frame_op(q)
    Quantity[...](Array([-8121.972, 0. , 20.8 ], dtype=float32), unit='pc')

    >>> p = u.Quantity([0., 0, 0], "km/s")

    >>> newq, newp = frame_op(q, p)
    >>> print(newq, newp, sep="\n")
    Quantity['length'](Array([-8121.972, 0. , 20.8  ], dtype=float32), unit='pc')
    Quantity['speed'](Array([ 12.9 , 245.6 , 7.78], dtype=float32), unit='km / s')

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "pc")
    >>> p = cx.CartesianVel3D.from_([0, 0, 0], "km/s")

    >>> newq, newp = frame_op(q, p)
    >>> print(newq, newp, sep="\n")
    <CartesianPos3D (x[pc], y[pc], z[pc])
        [-8121.972     0.       20.8  ]>
    <CartesianVel3D (d_x[km / s], d_y[km / s], d_z[km / s])
        [ 12.9  245.6    7.78]>

    """  # noqa: E501
    # rotation matrix to align x(ICRS) with the vector to the Galactic center
    rot_lat = GalileanRotation.from_euler("y", to_frame.galcen.lat)
    rot_lon = GalileanRotation.from_euler("z", -to_frame.galcen.lon)
    # extra roll away from the Galactic x-z plane
    roll = GalileanRotation.from_euler("x", to_frame.roll - to_frame.roll0)
    # construct transformation matrix
    R = (roll @ rot_lat @ rot_lon).simplify()

    # Translation by Sun-Galactic center distance around x' and rotate about y'
    # to account for tilt due to Sun's height above the plane
    z_d = u.ustrip("", to_frame.z_sun / to_frame.galcen.distance)  # [radian]
    H = GalileanRotation.from_euler("y", u.Quantity(jnp.asin(z_d), "rad"))

    # Post-rotation spatial offset to Galactic center.
    offset_q = GalileanSpatialTranslation(
        -to_frame.galcen.distance * jnp.asarray([1, 0, 0])
    )

    # Post-rotation velocity offset
    offset_v = VelocityBoost(to_frame.galcen_v_sun)

    # Total Operator
    return R | offset_q | H | offset_v


# ---------------------------------------------------------------


@dispatch
def frame_transform_op(from_frame: Galactocentric, to_frame: ICRS, /) -> Pipe:
    r"""Return a Galactocentric to ICRS frame transformation operator.

    Examples
    --------
    For this example we compare against the Astropy implementation of the
    frame transformation:

    >>> import unxt as u
    >>> import coordinax as cx
    >>> import astropy.coordinates as apyc

    The location of Vega in Galactocentric coordinates:

    >>> apy_gcf = apyc.Galactocentric()
    >>> apy_gcf
    <Galactocentric Frame (galcen_coord=<ICRS Coordinate: (ra, dec) in deg
        (266.4051, -28.936175)>, galcen_distance=8.122 kpc, galcen_v_sun=(12.9, 245.6, 7.78) km / s, z_sun=20.8 pc, roll=0.0 deg)>

    >>> vega = apyc.SkyCoord(
    ...     ra=279.23473479 * u.unit("deg"), dec=38.78368896 * u.unit("deg"),
    ...     distance=25 * u.unit("pc"),
    ...     pm_ra_cosdec=200 * u.unit("mas / yr"), pm_dec=-286 * u.unit("mas / yr"),
    ...     radial_velocity=-13.9 * u.unit("km / s")
    ... ).transform_to(apy_gcf)
    >>> print(vega)
    <SkyCoord (Galactocentric: ...): (x, y, z) in pc
        (-8112.89970167, 21.79911216, 29.01384942)
     (v_x, v_y, v_z) in km / s
        (34.06711868, 234.61647066, -28.75976702)>

    Transforming to an ICRS frame:

    >>> vega.transform_to(apyc.ICRS())
    <SkyCoord (ICRS): (ra, dec, distance) in (deg, deg, pc)
        (279.23473479, 38.78368896, 25.)
     (pm_ra_cosdec, pm_dec, radial_velocity) in (mas / yr, mas / yr, km / s)
        (200., -286., -13.9)>

    Now we can use `coordinax` to perform the same transformation!

    Converting the Astropy objects to `coordinax`:

    >>> vega_q = cx.CartesianPos3D.from_(vega.galactocentric.data)
    >>> vega_p = cx.CartesianVel3D.from_(vega.galactocentric.data.differentials["s"])

    >>> icrs_frame = cx.frames.ICRS()
    >>> gcf_frame = cx.frames.Galactocentric.from_(apy_gcf)

    Define the transformation operator:

    >>> frame_op = cx.frames.frame_transform_op(gcf_frame, icrs_frame)

    Apply the transformation:

    >>> vega_icrs_q, vega_icrs_p = frame_op(vega_q, vega_p)

    >>> vega_icrs_q = vega_icrs_q.vconvert(cx.vecs.LonLatSphericalPos)
    >>> vega_icrs_p = vega_icrs_p.vconvert(cx.vecs.LonCosLatSphericalVel, vega_icrs_q)
    >>> print(vega_icrs_q.uconvert({u.dimension("angle"): "deg", u.dimension("length"): "pc"}))
    <LonLatSphericalPos (lon[deg], lat[deg], distance[pc])
        [279.235  38.784  25.   ]>
    >>> print(vega_icrs_p.uconvert({u.dimension("angular speed"): "mas / yr", u.dimension("speed"): "km/s"}))
    <LonCosLatSphericalVel (d_lon_coslat[mas / yr], d_lat[mas / yr], d_distance[km / s])
        [ 200.001 -286.  -13.9  ]>

    It matches!

    Let's do it again for a few different input types:

    >>> q = u.Quantity([0, 0, 0], "pc")
    >>> frame_op(q).round(0)
    Quantity['length'](Array([ -446., -7094., -3930.], dtype=float32), unit='pc')

    >>> p = u.Quantity([0., 0, 0], "km/s")

    >>> newq, newp = frame_op(q, p)
    >>> print(newq.round(0), newp.round(0), sep="\n")
    Quantity['length'](Array([ -446., -7094., -3930.], dtype=float32), unit='pc')
    Quantity['speed'](Array([-114., 122., -181.], dtype=float32), unit='km / s')

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "pc")
    >>> p = cx.CartesianVel3D.from_([0, 0, 0], "km/s")

    >>> newq, newp = frame_op(q, p)
    >>> print(newq, newp, sep="\n")
    <CartesianPos3D (x[pc], y[pc], z[pc])
        [ -445.689 -7094.056 -3929.708]>
    <CartesianVel3D (d_x[km / s], d_y[km / s], d_z[km / s])
        [-113.868  122.047 -180.79 ]>

    """  # noqa: E501
    icrs2gcf = frame_transform_op(to_frame, from_frame)  # pylint: disable=W1114
    return icrs2gcf.inverse.simplify()



"""Representation of coordinates in different systems."""

__all__ = ["ToUnitsOptions"]

from enum import Enum


class ToUnitsOptions(Enum):
    """Options for the units argument of `AbstractVector.uconvert`."""

    consistent = "consistent"
    """Convert to consistent units."""



"""Base classes for operators on coordinates and potentials."""

__all__: list[str] = []

import functools

from plum import dispatch

from .base import AbstractOperator
from .identity import Identity
from .pipe import Pipe


@dispatch(precedence=-1)  # very low priority
def simplify_op(op: AbstractOperator, /) -> AbstractOperator:
    """Return the operator unchanged.

    Examples
    --------
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 0, 0], "m")
    >>> cx.ops.simplify_op(op) is op
    True

    """
    return op


@dispatch  # very low priority
def simplify_op(op: Identity, /) -> Identity:
    """Return the operator unchanged.

    Examples
    --------
    >>> import coordinax as cx

    >>> op = cx.ops.Identity()
    >>> cx.ops.simplify_op(op) is op
    True

    """
    return op


@dispatch
def simplify_op(seq: Pipe, /) -> AbstractOperator:
    """Simplify a sequence of Operators.

    This simplifies the sequence of operators by removing any that reduce to
    the Identity operator.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> shift = cx.ops.GalileanSpatialTranslation(u.Quantity([1, 2, 3], "km"))
    >>> boost = cx.ops.GalileanBoost(u.Quantity([10, 20, 30], "km/s"))

    >>> seq = shift | cx.ops.Identity() | boost
    >>> seq
    Pipe((
        GalileanSpatialTranslation(...), Identity(), GalileanBoost(...)
    ))

    >>> cx.ops.simplify_op(seq)
    Pipe(( GalileanSpatialTranslation(...), GalileanBoost(...) ))

    """
    # TODO: more sophisticated operator fusion. This just applies pair-wise
    # simplification.
    return functools.reduce(simplify_op, seq.operators, Identity())


# ======================================================================
# 2-Operator Simplifications


@dispatch(precedence=-1)
def simplify_op(op1: AbstractOperator, op2: AbstractOperator, /) -> Pipe:
    """Simplify two operators into a sequence.

    Examples
    --------
    >>> import coordinax as cx

    >>> op1 = cx.ops.GalileanSpatialTranslation.from_([1, 0, 0], "m")
    >>> op2 = cx.ops.GalileanBoost.from_([0, 1, 0], "m/s")
    >>> cx.ops.simplify_op(op1, op2)
    Pipe((
        GalileanSpatialTranslation(CartesianPos3D( ... )),
        GalileanBoost(CartesianVel3D( ... ))
    ))

    """
    return Pipe((op1, op2))


@dispatch(precedence=1)
def simplify_op(op1: AbstractOperator, op2: Identity) -> AbstractOperator:
    """Simplify an operator with the identity.

    Examples
    --------
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 0, 0], "m")
    >>> cx.ops.simplify_op(op, cx.ops.Identity())
    GalileanSpatialTranslation(...)

    """
    return op1


@dispatch
def simplify_op(op1: Identity, op2: AbstractOperator) -> AbstractOperator:
    """Simplify an operator with the identity.

    Examples
    --------
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 0, 0], "m")
    >>> cx.ops.simplify_op(cx.ops.Identity(), op)
    GalileanSpatialTranslation(...)

    """
    return op2


@dispatch
def simplify_op(op1: Pipe, op2: AbstractOperator) -> Pipe:
    """Simplify two sequences of operators by concatenating them.

    Examples
    --------
    >>> import coordinax as cx

    >>> sop1 = cx.ops.Pipe((cx.ops.Identity(), cx.ops.Identity()))
    >>> sop2 = cx.ops.Pipe((cx.ops.Identity(),))
    >>> cx.ops.simplify_op(sop1, sop2)
    Pipe((Identity(), Identity(), Identity()))

    """
    return op1 | op2


@dispatch
def simplify_op(op1: Pipe, op2: Pipe) -> Pipe:
    """Simplify two sequences of operators by concatenating them.

    Examples
    --------
    >>> import coordinax as cx

    >>> sop1 = cx.ops.Pipe((cx.ops.Identity(), cx.ops.Identity()))
    >>> sop2 = cx.ops.Pipe((cx.ops.Identity(),))
    >>> cx.ops.simplify_op(sop1, sop2)
    Pipe((Identity(), Identity(), Identity()))

    """
    return Pipe(op1.operators + op2.operators)



"""Built-in vector classes."""

__all__ = [
    "CartesianAcc3D",
    "CartesianPos3D",
    "CartesianVel3D",
]

from dataclasses import replace
from functools import partial
from typing import Any, final
from typing_extensions import override

import equinox as eqx
import jax
from jaxtyping import ArrayLike
from plum import dispatch
from quax import register

import quaxed.lax as qlax
import quaxed.numpy as jnp
import unxt as u
from unxt.quantity import AbstractQuantity

import coordinax._src.typing as ct
from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D
from .generic import CartesianGeneric3D
from coordinax._src.distances import BatchableLength
from coordinax._src.utils import classproperty, is_any_quantity
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.mixins import AvalMixin

#####################################################################
# Position


@final
class CartesianPos3D(AbstractPos3D):
    """Cartesian 3D Position.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_(u.Quantity([1, 2, 3], "m"))
    >>> print(vec)
    <CartesianPos3D (x[m], y[m], z[m])
        [1 2 3]>

    """

    x: BatchableLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""X coordinate :math:`x \in (-\infty,+\infty)`."""

    y: BatchableLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""Y coordinate :math:`y \in (-\infty,+\infty)`."""

    z: BatchableLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""Z coordinate :math:`z \in (-\infty,+\infty)`."""

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["CartesianVel3D"]:
        """Return the differential class.

        Examples
        --------
        >>> import coordinax as cx
        >>> print(cx.vecs.CartesianPos3D.differential_cls)
        <class 'coordinax...CartesianVel3D'>

        """
        return CartesianVel3D


# =====================================================
# Primitives


@register(jax.lax.add_p)  # type: ignore[misc]
def _add_cart3d_pos(lhs: CartesianPos3D, rhs: AbstractPos, /) -> CartesianPos3D:
    """Subtract two vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> s = cx.SphericalPos(r=u.Quantity(1, "km"), theta=u.Quantity(90, "deg"),
    ...                     phi=u.Quantity(0, "deg"))
    >>> print(q + s)
    <CartesianPos3D (x[km], y[km], z[km])
        [2. 2. 3.]>

    """
    cart = rhs.vconvert(CartesianPos3D)
    return jax.tree.map(jnp.add, lhs, cart, is_leaf=is_any_quantity)


# ------------------------------------------------
# Dot product
# TODO: see implementation in https://github.com/google/tree-math for how to do
# this more generally.


@register(jax.lax.dot_general_p)  # type: ignore[misc]
def _dot_general_cart3d(
    lhs: CartesianPos3D, rhs: CartesianPos3D, /, **kwargs: Any
) -> AbstractQuantity:
    """Dot product of two vectors.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> q1 = cx.vecs.CartesianPos3D.from_([1, 2, 3], "m")
    >>> q2 = cx.vecs.CartesianPos3D.from_([4, 5, 6], "m")

    >>> jnp.dot(q1, q2)
    Quantity['area'](Array(32, dtype=int32), unit='m2')

    """
    return lhs.x * rhs.x + lhs.y * rhs.y + lhs.z * rhs.z


# ------------------------------------------------


@register(jax.lax.neg_p)  # type: ignore[misc]
def _neg_p_cart3d_pos(obj: CartesianPos3D, /) -> CartesianPos3D:
    """Negate the `coordinax.CartesianPos3D`.

    Examples
    --------
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> print(-q)
    <CartesianPos3D (x[km], y[km], z[km])
        [-1 -2 -3]>

    """
    return jax.tree.map(qlax.neg, obj)


@register(jax.lax.sub_p)  # type: ignore[misc]
def _sub_cart3d_pos(lhs: CartesianPos3D, rhs: AbstractPos, /) -> CartesianPos3D:
    """Subtract two vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> s = cx.SphericalPos(r=u.Quantity(1, "km"), theta=u.Quantity(90, "deg"),
    ...                     phi=u.Quantity(0, "deg"))
    >>> print(q - s)
    <CartesianPos3D (x[km], y[km], z[km])
        [0. 2. 3.]>

    """
    cart = rhs.vconvert(CartesianPos3D)
    return jax.tree.map(jnp.subtract, lhs, cart)


# =====================================================
# Functions


# from coordinax.vectors.funcs
@dispatch
@partial(eqx.filter_jit, inline=True)
def normalize_vector(obj: CartesianPos3D, /) -> CartesianGeneric3D:
    """Return the norm of the vector.

    This has length 1.

    .. note::

        The unit vector is dimensionless, even if the input vector has units.
        This is because the unit vector is a ratio of two quantities: each
        component and the norm of the vector.

    Returns
    -------
    CartesianGeneric3D
        The norm of the vector.

    Examples
    --------
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> print(cx.vecs.normalize_vector(q))
    <CartesianGeneric3D (x[], y[], z[])
        [0.267 0.535 0.802]>

    """
    norm = obj.norm()
    return CartesianGeneric3D(x=obj.x / norm, y=obj.y / norm, z=obj.z / norm)


#####################################################################
# Velocity


@final
class CartesianVel3D(AvalMixin, AbstractVel3D):
    """Cartesian 3D Velocity.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.CartesianVel3D.from_([1, 2, 3], "m/s")
    >>> print(vec)
    <CartesianVel3D (d_x[m / s], d_y[m / s], d_z[m / s])
        [1 2 3]>

    """

    d_x: ct.BatchableSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""X speed :math:`dx/dt \in [-\infty, \infty]."""

    d_y: ct.BatchableSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Y speed :math:`dy/dt \in [-\infty, \infty]."""

    d_z: ct.BatchableSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Z speed :math:`dz/dt \in [-\infty, \infty]."""

    @override
    @classproperty
    @classmethod
    def integral_cls(cls) -> type[CartesianPos3D]:
        return CartesianPos3D

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["CartesianAcc3D"]:
        """Return the differential class.

        Examples
        --------
        >>> import coordinax as cx
        >>> print(cx.vecs.CartesianVel3D.differential_cls)
        <class 'coordinax...CartesianAcc3D'>

        """
        return CartesianAcc3D

    @partial(eqx.filter_jit, inline=True)
    def norm(self, _: AbstractPos3D | None = None, /) -> ct.BatchableSpeed:
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> c = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
        >>> c.norm()
        Quantity['speed'](Array(3.7416575, dtype=float32), unit='km / s')

        """
        return jnp.sqrt(self.d_x**2 + self.d_y**2 + self.d_z**2)


# -----------------------------------------------------
# Method dispatches


@register(jax.lax.add_p)  # type: ignore[misc]
def _add_pp(lhs: CartesianVel3D, rhs: CartesianVel3D, /) -> CartesianVel3D:
    """Add two Cartesian velocities.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> print(q + q)
    <CartesianVel3D (d_x[km / s], d_y[km / s], d_z[km / s])
        [2 4 6]>

    """
    return jax.tree.map(jnp.add, lhs, rhs)


@register(jax.lax.sub_p)  # type: ignore[misc]
def _sub_v3_v3(lhs: CartesianVel3D, other: CartesianVel3D, /) -> CartesianVel3D:
    """Subtract two differentials.

    Examples
    --------
    >>> from coordinax import CartesianPos3D, CartesianVel3D
    >>> q = CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> print(q - q)
    <CartesianVel3D (d_x[km / s], d_y[km / s], d_z[km / s])
        [0 0 0]>

    """
    return jax.tree.map(jnp.subtract, lhs, other)


#####################################################################
# Acceleration


@final
class CartesianAcc3D(AvalMixin, AbstractAcc3D):
    """Cartesian differential representation."""

    d2_x: ct.BatchableAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""X acceleration :math:`d^2x/dt^2 \in [-\infty, \infty]."""

    d2_y: ct.BatchableAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""Y acceleration :math:`d^2y/dt^2 \in [-\infty, \infty]."""

    d2_z: ct.BatchableAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""Z acceleration :math:`d^2z/dt^2 \in [-\infty, \infty]."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[CartesianVel3D]:
        return CartesianVel3D

    # -----------------------------------------------------
    # Methods

    @override
    @partial(jax.jit, inline=True)
    def norm(
        self, _: AbstractVel3D | None = None, __: AbstractPos3D | None = None, /
    ) -> ct.BatchableAcc:
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> c = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "km/s2")
        >>> c.norm()
        Quantity['acceleration'](Array(3.7416575, dtype=float32), unit='km / s2')

        """
        return jnp.sqrt(self.d2_x**2 + self.d2_y**2 + self.d2_z**2)


# -----------------------------------------------------
# Method dispatches


@register(jax.lax.add_p)  # type: ignore[misc]
def _add_aa(lhs: CartesianAcc3D, rhs: CartesianAcc3D, /) -> CartesianAcc3D:
    """Add two Cartesian accelerations.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> print(q + q)
    <CartesianAcc3D (d2_x[km / s2], d2_y[km / s2], d2_z[km / s2])
        [2 4 6]>

    """
    return jax.tree.map(jnp.add, lhs, rhs)


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_ac3(lhs: ArrayLike, rhs: CartesianPos3D, /) -> CartesianPos3D:
    """Scale a position by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    >>> v = cx.CartesianPos3D.from_([1, 2, 3], "km")

    >>> print(2 * v)
    <CartesianPos3D (x[km], y[km], z[km])
        [2 4 6]>

    >>> print(jnp.multiply(2, v))
    <CartesianPos3D (x[km], y[km], z[km])
        [2 4 6]>

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )

    # Scale the components
    return replace(rhs, x=lhs * rhs.x, y=lhs * rhs.y, z=lhs * rhs.z)


@register(jax.lax.sub_p)  # type: ignore[misc]
def _sub_a3_a3(lhs: CartesianAcc3D, rhs: CartesianAcc3D, /) -> CartesianAcc3D:
    """Subtract two accelerations.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> print(q - q)
    <CartesianAcc3D (d2_x[km / s2], d2_y[km / s2], d2_z[km / s2])
        [0 0 0]>

    """
    return jax.tree.map(jnp.subtract, lhs, rhs)



"""Position Vector ABC.

This is Private API.

"""

__all__ = ["AbstractPos", "POSITION_CLASSES"]

from .core import POSITION_CLASSES, AbstractPos

# Register by import
# isort: split
from . import (
    register_convert,  # noqa: F401
    register_primitives,  # noqa: F401
)



"""3-dimensional constructors."""
# pylint: disable=duplicate-code

__all__: list[str] = []

from plum import dispatch

from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D

#####################################################################


@dispatch(precedence=1)
def vector(cls: type[AbstractPos3D], obj: AbstractPos3D, /) -> AbstractPos3D:
    """Construct from a 3D position.

    Examples
    --------
    >>> import coordinax as cx

    >>> cart = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> cx.vecs.AbstractPos3D.from_(cart) is cart
    True

    >>> sph = cart.vconvert(cx.SphericalPos)
    >>> cx.vecs.AbstractPos3D.from_(sph) is sph
    True

    >>> cyl = cart.vconvert(cx.vecs.CylindricalPos)
    >>> cx.vecs.AbstractPos3D.from_(cyl) is cyl
    True

    """
    return obj


#####################################################################


@dispatch(precedence=1)
def vector(cls: type[AbstractVel3D], obj: AbstractVel3D, /) -> AbstractVel3D:
    """Construct from a 3D velocity.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.CartesianPos3D.from_([1, 1, 1], "km")

    >>> cart = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> cx.vecs.AbstractVel3D.from_(cart) is cart
    True

    >>> sph = cart.vconvert(cx.SphericalVel, q)
    >>> cx.vecs.AbstractVel3D.from_(sph) is sph
    True

    >>> cyl = cart.vconvert(cx.vecs.CylindricalVel, q)
    >>> cx.vecs.AbstractVel3D.from_(cyl) is cyl
    True

    """
    return obj


#####################################################################


@dispatch(precedence=1)
def vector(cls: type[AbstractAcc3D], obj: AbstractAcc3D, /) -> AbstractAcc3D:
    """Construct from a 3D velocity.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.CartesianPos3D.from_([1, 1, 1], "km")
    >>> p = cx.CartesianVel3D.from_([1, 1, 1], "km/s")

    >>> cart = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> cx.vecs.AbstractAcc3D.from_(cart) is cart
    True

    >>> sph = cart.vconvert(cx.vecs.SphericalAcc, p, q)
    >>> cx.vecs.AbstractAcc3D.from_(sph) is sph
    True

    >>> cyl = cart.vconvert(cx.vecs.CylindricalAcc, p, q)
    >>> cx.vecs.AbstractAcc3D.from_(cyl) is cyl
    True

    """
    return obj



"""Representation of coordinates in different systems."""

__all__ = ["AbstractPos"]

from dataclasses import replace
from typing import Any

import equinox as eqx
import jax
from jaxtyping import ArrayLike
from plum import convert
from quax import quaxify, register

import quaxed.lax as qlax
import quaxed.numpy as jnp
import unxt as u
from dataclassish import field_items
from unxt.quantity import AbstractQuantity

from .core import AbstractPos
from coordinax._src.vectors.api import vconvert
from coordinax._src.vectors.base import AttrFilter


@register(jax.lax.add_p)  # type: ignore[misc]
def _add_qq(lhs: AbstractPos, rhs: AbstractPos, /) -> AbstractPos:
    # The base implementation is to convert to Cartesian and perform the
    # operation.  Cartesian coordinates do not have any branch cuts or
    # singularities or ranges that need to be handled, so this is a safe
    # default.
    cart_cls = lhs._cartesian_cls  # noqa: SLF001
    cart_cls = eqx.error_if(
        cart_cls,
        isinstance(lhs, cart_cls) and isinstance(rhs, cart_cls),
        "must register a Cartesian-specific dispatch for {cart_cls} addition",
    )
    clhs = lhs.vconvert(cart_cls)
    crhs = rhs.vconvert(cart_cls)
    return (clhs + crhs).vconvert(type(lhs))


# ------------------------------------------------
# Dot product
# TODO: see implementation in https://github.com/google/tree-math for how to do
# this more generally.


@register(jax.lax.dot_general_p)  # type: ignore[misc]
def _dot_general_pos(
    lhs: AbstractPos, rhs: AbstractPos, /, **kwargs: Any
) -> AbstractQuantity:
    """Dot product of two vectors.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.vecs.SphericalPos(
    ...     r=u.Quantity([1, 2, 3], "m"),
    ...     theta=u.Quantity([0, 0, 0], "rad"),
    ...     phi=u.Quantity([0, 0, 0], "rad"))

    >>> jnp.dot(vec, vec)
    Quantity['area'](Array([1., 4., 9.], dtype=float32), unit='m2')

    """
    return qlax.dot_general(
        lhs.vconvert(lhs._cartesian_cls),  # noqa: SLF001
        rhs.vconvert(lhs._cartesian_cls),  # noqa: SLF001
        **kwargs,
    )


# ------------------------------------------------


@register(jax.lax.div_p)  # type: ignore[misc]
def _div_pos_v(lhs: AbstractPos, rhs: ArrayLike) -> AbstractPos:
    """Divide a vector by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "m")
    >>> jnp.divide(vec, 2).x
    Quantity['length'](Array(0.5, dtype=float32), unit='m')

    >>> (vec / 2).x
    Quantity['length'](Array(0.5, dtype=float32), unit='m')

    """
    return replace(
        lhs, **{k: jnp.divide(v, rhs) for k, v in field_items(AttrFilter, lhs)}
    )


# ------------------------------------------------


@register(jax.lax.eq_p)  # type: ignore[misc]
def _eq_pos_pos(lhs: AbstractPos, rhs: AbstractPos, /) -> ArrayLike:
    """Element-wise equality of two positions."""
    return lhs == rhs


# ------------------------------------------------


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_v_pos(lhs: ArrayLike, rhs: AbstractPos, /) -> AbstractPos:
    """Scale a position by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "m")
    >>> print(jnp.multiply(2, vec))
    <CartesianPos3D (x[m], y[m], z[m])
        [2 4 6]>

    Most of the position classes have specific dispatches for this operation.
    So let's define a new class and try it out:

    >>> from typing import ClassVar
    >>> class MyCartesian(cx.vecs.AbstractPos):
    ...     x: u.Quantity
    ...     y: u.Quantity
    ...     z: u.Quantity
    ...     _dimensionality: ClassVar[int] = 3
    ...
    >>> MyCartesian._cartesian_cls = MyCartesian  # hack

    Add conversion to Quantity:

    >>> from plum import conversion_method
    >>> @conversion_method(MyCartesian, u.Quantity)
    ... def _to_quantity(x: MyCartesian, /) -> u.Quantity:
    ...     return jnp.stack((x.x, x.y, x.z), axis=-1)

    Add representation transformation

    >>> from plum import dispatch
    >>> @dispatch
    ... def vconvert(target: type[MyCartesian], current: MyCartesian, /) -> MyCartesian:
    ...     return current

    >>> vec = MyCartesian(x=u.Quantity([1], "m"),
    ...                   y=u.Quantity([2], "m"),
    ...                   z=u.Quantity([3], "m"))

    First hit the non-scalar error:

    >>> try: jnp.multiply(jnp.asarray([[1, 1, 1]]), vec)
    ... except Exception as e: print(e)
    must be a scalar, not <class 'jaxlib.xla_extension.ArrayImpl'>

    Then hit the Cartesian-specific dispatch error:

    >>> try: jnp.multiply(2, vec)
    ... except Exception as e: print(e)
    must register a Cartesian-specific dispatch

    Now a real example. For this we need to define the Cartesian-specific
    dispatches:

    >>> MyCartesian._cartesian_cls = cx.CartesianPos3D
    >>> @dispatch
    ... def vconvert(target: type[cx.CartesianPos3D],current: MyCartesian, /) -> cx.CartesianPos3D:
    ...     return cx.CartesianPos3D(x=current.x, y=current.y, z=current.z)
    >>> @dispatch
    ... def vconvert(target: type[MyCartesian], current: cx.CartesianPos3D, /) -> MyCartesian:
    ...     return MyCartesian(x=current.x, y=current.y, z=current.z)

    >>> print(jnp.multiply(2, vec))
    <MyCartesian (x[m], y[m], z[m])
        [[2 4 6]]>

    """  # noqa: E501
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )
    rhs = eqx.error_if(
        rhs,
        isinstance(rhs, rhs._cartesian_cls),  # noqa: SLF001
        "must register a Cartesian-specific dispatch",
    )

    rc = rhs.vconvert(rhs._cartesian_cls)  # noqa: SLF001
    nr = qlax.mul(lhs, rc)
    return nr.vconvert(type(rhs))


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_pos_v(lhs: AbstractPos, rhs: ArrayLike, /) -> AbstractPos:
    """Scale a position by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "m")
    >>> print(jnp.multiply(vec, 2))
    <CartesianPos3D (x[m], y[m], z[m])
        [2 4 6]>

    """
    return qlax.mul(rhs, lhs)  # re-dispatch on the other side


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_pos_pos(lhs: AbstractPos, rhs: AbstractPos, /) -> u.Quantity:
    """Multiply two positions.

    This is required to take the dot product of two vectors.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D(
    ...     x=u.Quantity([1, 2, 3], "m"),
    ...     y=u.Quantity([4, 5, 6], "m"),
    ...     z=u.Quantity([7, 8, 9], "m"))

    >>> jnp.multiply(vec, vec)  # element-wise multiplication
    Quantity['area'](Array([[ 1, 16, 49],
                            [ 4, 25, 64],
                            [ 9, 36, 81]], dtype=int32), unit='m2')

    >>> jnp.linalg.vector_norm(vec, axis=-1)
    Quantity['length'](Array([ 8.124039,  9.643651, 11.224972], dtype=float32), unit='m')

    """  # noqa: E501
    lq: u.Quantity = convert(lhs.vconvert(lhs._cartesian_cls), u.Quantity)  # noqa: SLF001
    rq: u.Quantity = convert(rhs.vconvert(rhs._cartesian_cls), u.Quantity)  # noqa: SLF001
    return qlax.mul(lq, rq)  # re-dispatch to Quantities


# ------------------------------------------------


@register(jax.lax.neg_p)  # type: ignore[misc]
def _neg_pos(obj: AbstractPos, /) -> AbstractPos:
    """Negate the vector.

    The default implementation is to go through Cartesian coordinates.

    Examples
    --------
    >>> import coordinax as cx
    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "m")
    >>> print(-vec)
    <CartesianPos3D (x[m], y[m], z[m])
        [-1 -2 -3]>

    """
    cart = vconvert(obj._cartesian_cls, obj)  # noqa: SLF001
    negcart = jnp.negative(cart)
    return vconvert(type(obj), negcart)


# ------------------------------------------------


@register(jax.lax.reshape_p)  # type: ignore[misc]
def _reshape_pos(
    operand: AbstractPos, *, new_sizes: tuple[int, ...], **kwargs: Any
) -> AbstractPos:
    """Reshape the components of the vector.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> import quaxed.numpy as jnp

    >>> vec = cx.CartesianPos3D(x=u.Quantity([1, 2, 3], "m"),
    ...                         y=u.Quantity([4, 5, 6], "m"),
    ...                         z=u.Quantity([7, 8, 9], "m"))
    >>> vec = jnp.reshape(vec, shape=(3, 1, 3))  # (n_components *shape)
    >>> print(vec)
    <CartesianPos3D (x[m], y[m], z[m])
        [[[[1 4 7]
           [2 5 8]
           [3 6 9]]]]>

    """
    # Adjust the sizes for the components
    new_sizes = (new_sizes[0] // len(operand.components), *new_sizes[1:])
    # TODO: check integer division
    # Reshape the components
    return replace(
        operand,
        **{
            k: quaxify(jax.lax.reshape_p.bind)(v, new_sizes=new_sizes, **kwargs)
            for k, v in field_items(operand)
        },
    )


# ------------------------------------------------


@register(jax.lax.sub_p)  # type: ignore[misc]
def _sub_qq(lhs: AbstractPos, rhs: AbstractPos) -> AbstractPos:
    """Add another object to this vector."""
    # The base implementation is to convert to Cartesian and perform the
    # operation.  Cartesian coordinates do not have any branch cuts or
    # singularities or ranges that need to be handled, so this is a safe
    # default.
    return qlax.sub(
        lhs.vconvert(lhs._cartesian_cls),  # noqa: SLF001
        rhs.vconvert(lhs._cartesian_cls),  # noqa: SLF001
    ).vconvert(type(lhs))



"""`coordinax.angle` private module.

Note that this module is private. Users should use the public API.

This module depends on the following modules:

- utils & typing

"""

__all__ = ["AbstractAngle", "Angle", "BatchableAngle", "BatchableAngleQ"]

from .angle import Angle
from .base import AbstractAngle
from .typing import BatchableAngle, BatchableAngleQ

# isort: split
# Register the dispatching
from . import register_converters, register_primitives  # noqa: F401



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from collections.abc import Mapping
from typing import Any

from jaxtyping import ArrayLike
from plum import dispatch

import quaxed.numpy as jnp
import unxt as u
from dataclassish import field_items
from unxt.quantity import AbstractQuantity

from .vector import AbstractVector
from coordinax._src.typing import Unit


@dispatch
def vector(obj: AbstractVector, /) -> AbstractVector:
    """Construct a vector from a vector.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> cart = cx.vecs.CartesianPos2D.from_([1, 2], "km")
    >>> cx.vector(cart) is cart
    True

    """
    return obj


@dispatch
def vector(cls: type[AbstractVector], obj: Mapping[str, Any], /) -> AbstractVector:
    """Construct a vector from a mapping.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> xs = {"x": u.Quantity(1, "m"), "y": u.Quantity(2, "m"),
    ...       "z": u.Quantity(3, "m")}
    >>> vec = cx.CartesianPos3D.from_(xs)
    >>> print(vec)
    <CartesianPos3D (x[m], y[m], z[m])
        [1 2 3]>

    >>> xs = {"x": u.Quantity([1, 2], "m"), "y": u.Quantity([3, 4], "m"),
    ...       "z": u.Quantity([5, 6], "m")}
    >>> vec = cx.CartesianPos3D.from_(xs)
    >>> print(vec)
    <CartesianPos3D (x[m], y[m], z[m])
        [[1 3 5]
        [2 4 6]]>

    """
    return cls(**obj)


@dispatch
def vector(cls: type[AbstractVector], obj: AbstractQuantity, /) -> AbstractVector:
    """Construct a vector from a quantity.

    This will fail for most non-position vectors, except Cartesian vectors,
    since they generally do not have the same dimensions, nor can be converted
    from a Cartesian vector without additional information.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    Mismatch:

    >>> try: cx.vecs.CartesianPos1D.from_(u.Quantity([1, 2, 3], "m"))
    ... except ValueError as e: print(e)
    Cannot construct <class 'coordinax...CartesianPos1D'> from 3 components.

    Pos 1D:

    >>> cx.vecs.CartesianPos1D.from_(u.Quantity(1, "meter"))
    CartesianPos1D( x=Quantity[...](value=...i32[], unit=Unit("m")) )

    >>> cx.vecs.CartesianPos1D.from_(u.Quantity([1], "meter"))
    CartesianPos1D(x=Quantity[...](value=i32[], unit=Unit("m")))

    >>> cx.vecs.CartesianPos1D.from_(cx.Distance(1, "meter"))
    CartesianPos1D( x=Quantity[...](value=...i32[], unit=Unit("m")) )

    >>> cx.vecs.RadialPos.from_(u.Quantity(1, "meter"))
    RadialPos(r=Distance(value=...i32[], unit=Unit("m")))

    >>> cx.vecs.RadialPos.from_(u.Quantity([1], "meter"))
    RadialPos(r=Distance(value=...i32[], unit=Unit("m")))

    Vel 1D:

    >>> cx.vecs.CartesianVel1D.from_(u.Quantity(1, "m/s"))
    CartesianVel1D( d_x=Quantity[...]( value=...i32[], unit=Unit("m / s") ) )

    >>> cx.vecs.CartesianVel1D.from_(u.Quantity([1], "m/s"))
    CartesianVel1D( d_x=Quantity[...]( value=i32[], unit=Unit("m / s") ) )

    >>> cx.vecs.RadialVel.from_(u.Quantity(1, "m/s"))
    RadialVel( d_r=Quantity[...]( value=...i32[], unit=Unit("m / s") ) )

    >>> cx.vecs.RadialVel.from_(u.Quantity([1], "m/s"))
    RadialVel( d_r=Quantity[...]( value=i32[], unit=Unit("m / s") ) )

    Acc 1D:

    >>> cx.vecs.CartesianAcc1D.from_(u.Quantity(1, "m/s2"))
    CartesianAcc1D( d2_x=... )

    >>> cx.vecs.CartesianAcc1D.from_(u.Quantity([1], "m/s2"))
    CartesianAcc1D( d2_x=Quantity[...](value=i32[], unit=Unit("m / s2")) )

    >>> cx.vecs.RadialAcc.from_(u.Quantity(1, "m/s2"))
    RadialAcc( d2_r=... )

    >>> cx.vecs.RadialAcc.from_(u.Quantity([1], "m/s2"))
    RadialAcc( d2_r=Quantity[...](value=i32[], unit=Unit("m / s2")) )

    Pos 2D:

    >>> vec = cx.vecs.CartesianPos2D.from_(u.Quantity([1, 2], "m"))
    >>> vec
    CartesianPos2D(
        x=Quantity[...](value=i32[], unit=Unit("m")),
        y=Quantity[...](value=i32[], unit=Unit("m"))
    )

    Vel 2D:

    >>> vec = cx.vecs.CartesianVel2D.from_(u.Quantity([1, 2], "m/s"))
    >>> print(vec)
    <CartesianVel2D (d_x[m / s], d_y[m / s])
        [1 2]>

    Acc 2D:

    >>> vec = cx.vecs.CartesianAcc2D.from_(u.Quantity([1, 2], "m/s2"))
    >>> print(vec)
    <CartesianAcc2D (d2_x[m / s2], d2_y[m / s2])
        [1 2]>

    Pos 3D:

    >>> vec = cx.CartesianPos3D.from_(u.Quantity([1, 2, 3], "m"))
    >>> print(vec)
    <CartesianPos3D (x[m], y[m], z[m])
        [1 2 3]>

    Vel 3D:

    >>> vec = cx.CartesianVel3D.from_(u.Quantity([1, 2, 3], "m/s"))
    >>> print(vec)
    <CartesianVel3D (d_x[m / s], d_y[m / s], d_z[m / s])
        [1 2 3]>

    Acc 3D:

    >>> vec = cx.vecs.CartesianAcc3D.from_(u.Quantity([1, 2, 3], "m/s2"))
    >>> print(vec)
    <CartesianAcc3D (d2_x[m / s2], d2_y[m / s2], d2_z[m / s2])
        [1 2 3]>

    Generic 3D:

    >>> vec = cx.vecs.CartesianGeneric3D.from_(u.Quantity([1, 2, 3], "m"))
    >>> print(vec)
    <CartesianGeneric3D (x[m], y[m], z[m])
        [1 2 3]>

    """
    # Ensure the object is at least 1D
    obj = jnp.atleast_1d(obj)

    # Check the dimensions
    if obj.shape[-1] != cls._dimensionality():
        msg = f"Cannot construct {cls} from {obj.shape[-1]} components."
        raise ValueError(msg)

    # Map the components
    comps = {k: obj[..., i] for i, k in enumerate(cls.components)}

    # Construct the vector from the mapping
    return cls.from_(comps)


@dispatch
def vector(
    cls: type[AbstractVector], obj: ArrayLike | list[Any], unit: Unit | str, /
) -> AbstractVector:
    """Construct a vector from an array and unit.

    The ``ArrayLike[Any, (*#batch, N), "..."]`` is expected to have the
    components as the last dimension.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "meter")
    >>> print(vec)
    <CartesianPos3D (x[m], y[m], z[m])
        [1 2 3]>

    >>> xs = jnp.array([[1, 2, 3], [4, 5, 6]])
    >>> vec = cx.CartesianPos3D.from_(xs, "meter")
    >>> print(vec)
    <CartesianPos3D (x[m], y[m], z[m])
        [[1 2 3]
        [4 5 6]]>

    """
    obj = u.Quantity.from_(jnp.asarray(obj), unit)
    return cls.from_(obj)  # re-dispatch


@dispatch
def vector(cls: type[AbstractVector], obj: AbstractVector, /) -> AbstractVector:
    """Construct a vector from another vector.

    Raises
    ------
    TypeError
        If the object is not an instance of the vector class.

    Parameters
    ----------
    cls : type[AbstractVector], positional-only
        The vector class.
    obj : :class:`coordinax.AbstractVector`, positional-only
        The vector to construct from.

    Examples
    --------
    >>> import coordinax as cx

    Positions:

    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")

    >>> cart = cx.CartesianPos3D.from_(q)
    >>> print(cart)
    <CartesianPos3D (x[km], y[km], z[km])
        [1 2 3]>

    >>> cx.vecs.AbstractPos3D.from_(cart) is cart
    True

    >>> sph = cart.vconvert(cx.SphericalPos)
    >>> cx.vecs.AbstractPos3D.from_(sph) is sph
    True

    >>> cyl = cart.vconvert(cx.vecs.CylindricalPos)
    >>> cx.vecs.AbstractPos3D.from_(cyl) is cyl
    True

    Velocities:

    >>> p = cx.CartesianVel3D.from_([1, 2, 3], "km/s")

    >>> cart = cx.CartesianVel3D.from_(p)
    >>> cx.vecs.AbstractVel3D.from_(cart) is cart
    True

    >>> sph = cart.vconvert(cx.SphericalVel, q)
    >>> cx.vecs.AbstractVel3D.from_(sph) is sph
    True

    >>> cyl = cart.vconvert(cx.vecs.CylindricalVel, q)
    >>> cx.vecs.AbstractVel3D.from_(cyl) is cyl
    True

    Accelerations:

    >>> p = cx.CartesianVel3D.from_([1, 1, 1], "km/s")

    >>> cart = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> cx.vecs.AbstractAcc3D.from_(cart) is cart
    True

    >>> sph = cart.vconvert(cx.vecs.SphericalAcc, p, q)
    >>> cx.vecs.AbstractAcc3D.from_(sph) is sph
    True

    >>> cyl = cart.vconvert(cx.vecs.CylindricalAcc, p, q)
    >>> cx.vecs.AbstractAcc3D.from_(cyl) is cyl
    True

    """
    if not isinstance(obj, cls):
        msg = f"Cannot construct {cls} from {type(obj)}."
        raise TypeError(msg)

    # Avoid copying if the types are the same. `isinstance` is not strict
    # enough, so we use type() instead.
    if type(obj) is cls:  # pylint: disable=unidiomatic-typecheck
        return obj

    return cls(**dict(field_items(obj)))



"""Register primitives for AbstractVel."""

__all__: list[str] = []


import jax
from quax import register

import unxt as u
from dataclassish import field_items

from .core import AbstractVel
from coordinax._src.vectors.base_pos import AbstractPos

# ---------------------------------------------------------


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_vel_q(self: AbstractVel, other: u.Quantity["time"]) -> AbstractPos:
    """Multiply the vector by a time :class:`unxt.Quantity` to get a position.

    Examples
    --------
    >>> import quaxed.lax as qlax
    >>> import unxt as u
    >>> import coordinax as cx

    >>> dr = cx.vecs.RadialVel(u.Quantity(1, "m/s"))
    >>> vec = dr * u.Quantity(2, "s")
    >>> print(vec)
    <RadialPos (r[m])
        [2]>

    >>> print(qlax.mul(dr, u.Quantity(2, "s")))
    <RadialPos (r[m])
        [2]>

    """
    return self.integral_cls.from_({k[2:]: v * other for k, v in field_items(self)})



"""Base classes for operators on coordinates and potentials."""

__all__ = ["AbstractOperator"]

import textwrap
from abc import abstractmethod
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

import equinox as eqx
from jaxtyping import ArrayLike
from plum import dispatch

import unxt as u
from dataclassish import field_items, flags

from .api import simplify_op
from coordinax._src.vectors.base_pos import AbstractPos

if TYPE_CHECKING:
    from coordinax.ops import Pipe


class AbstractOperator(eqx.Module):  # type: ignore[misc]
    """Abstract base class for operators on coordinates and potentials.

    An operator is an object that defines a transformation on coordinates. It
    can be applied to a set of coordinates to produce a new set of coordinates.
    Operators can be composed together to form a sequence of transformations.

    When defining a new operator, it should be able to work on
    `coordinax.AbstractVector` objects. If it is a spatial operator, it should
    also be able to work on (`coordinax.AbstractPos`, `unxt.Quantity['time']`)
    pairs (and then also `coordinax.FourVector` objects). If the vector can be
    created from a `unxt.Quantity` object, then the operator should also be able
    to work on `unxt.Quantity` object.

    """

    # ---------------------------------------------------------------
    # Constructors

    @classmethod
    @dispatch(precedence=-1)
    def from_(
        cls: "type[AbstractOperator]", *args: object, **kwargs: object
    ) -> "AbstractOperator":
        """Construct from a set of arguments.

        This is a low-priority dispatch that will be called if no other
        dispatch is found. It just tries to pass the arguments to the
        constructor.

        """
        return cls(*args, **kwargs)

    @classmethod
    @dispatch
    def from_(
        cls: "type[AbstractOperator]", obj: Mapping[str, Any], /
    ) -> "AbstractOperator":
        """Construct from a mapping.

        Examples
        --------
        >>> import coordinax as cx
        >>> pipe = cx.ops.Identity() | cx.ops.Identity()
        >>> cx.ops.Pipe.from_({"operators": pipe})
        Pipe((Identity(), Identity()))

        """
        return cls(**obj)

    @classmethod
    @dispatch
    def from_(
        cls: "type[AbstractOperator]",
        x: ArrayLike | list[float | int],
        unit: str,  # TODO: support unit object
        /,
    ) -> "AbstractOperator":
        """Construct from a Quantity's value and unit.

        Examples
        --------
        >>> import coordinax as cx

        >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 1, 1], "km")
        >>> print(op.translation)
        <CartesianPos3D (x[km], y[km], z[km])
            [1 1 1]>

        >>> op = cx.ops.GalileanTranslation.from_([3e5, 1, 1, 1], "km")
        >>> print(op.translation)
        <FourVector (t[s], q=(x[km], y[km], z[km]))
            [1.001 1.    1.    1.   ]>

        >>> op = cx.ops.GalileanBoost.from_([1, 1, 1], "km/s")
        >>> print(op.velocity)
        <CartesianVel3D (d_x[km / s], d_y[km / s], d_z[km / s])
            [1 1 1]>

        """
        return cls(u.Quantity(x, unit))

    # ===========================================
    # Operator API

    @dispatch.abstract
    def __call__(
        self: "AbstractOperator",
        x: AbstractPos,  # noqa: ARG002
        /,
        **kwargs: Any,  # noqa: ARG002
    ) -> AbstractPos:
        """Apply the operator to the coordinates `x`."""
        msg = "implement this method in the subclass"
        raise TypeError(msg)

    @dispatch.abstract
    def __call__(
        self: "AbstractOperator",
        x: AbstractPos,  # noqa: ARG002
        t: u.Quantity["time"],  # noqa: ARG002
        /,
        **kwargs: Any,  # noqa: ARG002
    ) -> AbstractPos:
        """Apply the operator to the coordinates `x` at a time `t`."""
        msg = "implement this method in the subclass"
        raise TypeError(msg)

    # -------------------------------------------

    @property
    @abstractmethod
    def is_inertial(self) -> bool:
        """Whether the operation maintains an inertial reference frame."""
        ...

    @property
    @abstractmethod
    def inverse(self) -> "AbstractOperator":
        """The inverse of the operator."""
        ...

    def simplify(self) -> "AbstractOperator":
        """Simplify the operator.

        This method calls `coordinax.ops.simplify_op` to simplify the
        operator.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> op = cx.ops.Identity()
        >>> op.simplify() is op
        True

        >>> op = cx.ops.Pipe((cx.ops.Identity(), cx.ops.Identity()))
        >>> op.simplify()
        Identity()

        >>> op = cx.ops.GalileanOperator(translation=u.Quantity([0., 2., 3., 4.], "km"))
        >>> op.simplify()
        GalileanTranslation(FourVector( ... ))

        """
        return simplify_op(self)

    # ===========================================
    # Python stuff

    def __str__(self) -> str:
        """Return a string representation of the operator.

        Examples
        --------
        >>> import jax.numpy as jnp
        >>> import coordinax as cx

        >>> op = cx.ops.GalileanRotation([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        >>> print(op)
        GalileanRotation([[1 0 0]
         [0 1 0]
         [0 0 1]])

        >>> op = cx.ops.GalileanOperator(
        ...     translation=u.Quantity([0., 2, 3, 4], "km"),
        ...     velocity=u.Quantity([1., 2, 3], "km/s"),
        ...     rotation=jnp.eye(3).at[0, 2].set(1),
        ... )
        >>> print(op)
        GalileanOperator(
            rotation=GalileanRotation([[1. 0. 1.]
                [0. 1. 0.]
                [0. 0. 1.]]),
            translation=GalileanTranslation(<FourVector (t[s], q=(x[km], y[km], z[km]))
                    [0. 2. 3. 4.]>),
            velocity=GalileanBoost(<CartesianVel3D (d_x[km / s], d_y[km / s], d_z[km / s])
                    [1. 2. 3.]>)
        )

        """  # noqa: E501
        name = self.__class__.__name__
        fitems = field_items(flags.FilterRepr, self)
        if len(fitems) == 1:
            fs = str(fitems[0][1]).replace("\n", "\n    ")
        else:
            fs = ",\n".join(textwrap.indent(f"{k}={v}", "    ") for k, v in fitems)
            fs = "\n" + fs + "\n"

        return f"{name}({fs})"

    # ===========================================
    # Operator Composition

    def __or__(self, other: "AbstractOperator") -> "Pipe":
        """Compose with another operator.

        Examples
        --------
        >>> import coordinax as cx

        >>> op1 = cx.ops.GalileanRotation([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        >>> op2 = cx.ops.Identity()
        >>> op3 = op1 | op2
        >>> op3
        Pipe((GalileanRotation(rotation=i32[3,3]), Identity()))

        >>> op4 = cx.ops.Identity() | op3
        >>> op4
        Pipe((Identity(), GalileanRotation(rotation=i32[3,3]), Identity()))

        """
        from .pipe import Pipe

        if isinstance(other, Pipe):
            return other.__ror__(self)
        return Pipe((self, other))


@AbstractOperator.from_.dispatch  # type: ignore[attr-defined, misc]
def from_(cls: type[AbstractOperator], obj: AbstractOperator, /) -> AbstractOperator:
    """Construct an operator from another operator.

    Raises
    ------
    TypeError
        If the input object is not a subclass of the target class.

    Examples
    --------
    >>> import coordinax as cx

    If the object is the same type, it should return the object itself.

    >>> op = cx.ops.Identity()
    >>> cx.ops.Identity.from_(op) is op
    True

    If the object is a different type, it will error.

    >>> try:
    ...     cx.ops.GalileanBoost.from_(op)
    ... except TypeError as e:
    ...     print(e)
    Cannot construct <class 'coordinax...GalileanBoost'> from <class 'coordinax...Identity'>.

    Unless the object is a subclass of the target class.

    >>> class MyOperator(cx.ops.Identity):
    ...     pass

    >>> op = MyOperator()
    >>> op
    MyOperator()

    >>> newop = cx.ops.Identity.from_(op)
    >>> newop is op, isinstance(newop, cx.ops.Identity)
    (False, True)

    """  # noqa: E501
    if not isinstance(obj, cls):
        msg = f"Cannot construct {cls} from {type(obj)}."
        raise TypeError(msg)

    # avoid copying if the types are the same. Isinstance is not strict
    # enough, so we use type() instead.
    if type(obj) is cls:  # pylint: disable=unidiomatic-typecheck
        return obj

    return cls(**dict(field_items(obj)))



"""Base classes for angular quantities."""

__all__: list[str] = []


import equinox as eqx
from jaxtyping import Array, Shaped
from plum import add_promotion_rule

import unxt as u
from unxt._src.units.api import AbstractUnits
from unxt.quantity import AbstractQuantity

angle_dimension = u.dimension("angle")


class AbstractAngle(AbstractQuantity):  # type: ignore[misc]
    """Angular Quantity.

    See Also
    --------
    `coordinax.angle.Angle` : a concrete implementation of this class.

    Examples
    --------
    For this example, we will use the concrete implementation of
    `coordinax.angle.AbstractAngle`, `coordinax.angle.Angle`.

    >>> from coordinax.angle import Angle

    >>> a = Angle(90, "deg")
    >>> a
    Angle(Array(90, dtype=int32, weak_type=True), unit='deg')

    Angles have to have dimensions of angle.

    >>> try: Angle(90, "m")
    ... except ValueError as e: print(e)
    Angle must have units with angular dimensions.

    """

    value: eqx.AbstractVar[Shaped[Array, "*shape"]]
    """The value of the `AbstractQuantity`."""

    unit: eqx.AbstractVar[AbstractUnits]
    """The unit associated with this value."""

    def __check_init__(self) -> None:
        """Check the initialization."""
        if u.dimension_of(self) != angle_dimension:
            msg = f"{type(self).__name__} must have units with angular dimensions."
            raise ValueError(msg)


# Add a rule that when a AbstractAngle interacts with a Quantity, the
# angle degrades to a Quantity. This is necessary for many operations, e.g.
# division of an angle by non-dimensionless quantity where the resulting units
# are not those of an angle.
add_promotion_rule(AbstractAngle, u.Quantity, u.Quantity)
